"""
***************************************************************************
    EditScriptDialog.py
    ---------------------
    Date                 : December 2012
    Copyright            : (C) 2012 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "December 2012"
__copyright__ = "(C) 2012, Alexander Bruy"

import codecs
import inspect
import os
import traceback
import warnings

from qgis.core import (
    QgsApplication,
    QgsError,
    QgsFileUtils,
    QgsProcessingAlgorithm,
    QgsProcessingFeatureBasedAlgorithm,
    QgsSettings,
)
from qgis.gui import QgsCodeEditorWidget, QgsErrorDialog, QgsGui, QgsShortcutsManager
from qgis.processing import alg as algfactory
from qgis.PyQt import sip, uic
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QPalette
from qgis.PyQt.QtWidgets import QFileDialog, QMessageBox, QVBoxLayout
from qgis.utils import OverrideCursor, iface

from processing.gui.AlgorithmDialog import AlgorithmDialog
from processing.script import ScriptUtils

from .ScriptEdit import ScriptEdit

pluginPath = os.path.split(os.path.dirname(__file__))[0]

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    WIDGET, BASE = uic.loadUiType(os.path.join(pluginPath, "ui", "DlgScriptEditor.ui"))


class ScriptEditorDialog(BASE, WIDGET):
    hasChanged = False

    DIALOG_STORE = []

    def __init__(self, filePath=None, parent=None):
        super().__init__(parent)
        # SIP is totally messed up here -- the dialog wrapper or something
        # is always prematurely cleaned which results in broken QObject
        # connections throughout.
        # Hack around this by storing dialog instances in a global list to
        # prevent too early wrapper garbage collection
        ScriptEditorDialog.DIALOG_STORE.append(self)

        def clean_up_store():
            ScriptEditorDialog.DIALOG_STORE = [
                d for d in ScriptEditorDialog.DIALOG_STORE if d != self
            ]

        self.destroyed.connect(clean_up_store)

        self.setupUi(self)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        QgsGui.instance().enableAutoGeometryRestore(self)

        vl = QVBoxLayout()
        vl.setContentsMargins(0, 0, 0, 0)
        self.editor_container.setLayout(vl)

        self.editor = ScriptEdit()
        self.code_editor_widget = QgsCodeEditorWidget(self.editor)
        vl.addWidget(self.code_editor_widget)

        if iface is not None:
            self.toolBar.setIconSize(iface.iconSize())
            self.setStyleSheet(iface.mainWindow().styleSheet())

        self.actionOpenScript.setIcon(
            QgsApplication.getThemeIcon("/mActionScriptOpen.svg")
        )
        self.actionSaveScript.setIcon(
            QgsApplication.getThemeIcon("/mActionFileSave.svg")
        )
        self.actionSaveScriptAs.setIcon(
            QgsApplication.getThemeIcon("/mActionFileSaveAs.svg")
        )
        self.actionRunScript.setIcon(QgsApplication.getThemeIcon("/mActionStart.svg"))
        self.actionCut.setIcon(QgsApplication.getThemeIcon("/mActionEditCut.svg"))
        self.actionCopy.setIcon(QgsApplication.getThemeIcon("/mActionEditCopy.svg"))
        self.actionPaste.setIcon(QgsApplication.getThemeIcon("/mActionEditPaste.svg"))
        self.actionUndo.setIcon(QgsApplication.getThemeIcon("/mActionUndo.svg"))
        self.actionRedo.setIcon(QgsApplication.getThemeIcon("/mActionRedo.svg"))
        self.actionFindReplace.setIcon(
            QgsApplication.getThemeIcon("/mActionFindReplace.svg")
        )
        self.actionIncreaseFontSize.setIcon(
            QgsApplication.getThemeIcon("/mActionIncreaseFont.svg")
        )
        self.actionDecreaseFontSize.setIcon(
            QgsApplication.getThemeIcon("/mActionDecreaseFont.svg")
        )
        self.actionToggleComment.setIcon(
            QgsApplication.getThemeIcon(
                "console/iconCommentEditorConsole.svg",
                self.palette().color(QPalette.ColorRole.WindowText),
            )
        )
        QgsGui.shortcutsManager().initializeCommonAction(
            self.actionToggleComment, QgsShortcutsManager.CommonAction.CodeToggleComment
        )

        # Connect signals and slots
        self.actionOpenScript.triggered.connect(self.openScript)
        self.actionSaveScript.triggered.connect(self.save)
        self.actionSaveScriptAs.triggered.connect(self.saveAs)
        self.actionRunScript.triggered.connect(self.runAlgorithm)
        self.actionCut.triggered.connect(self.editor.cut)
        self.actionCopy.triggered.connect(self.editor.copy)
        self.actionPaste.triggered.connect(self.editor.paste)
        self.actionUndo.triggered.connect(self.editor.undo)
        self.actionRedo.triggered.connect(self.editor.redo)
        self.actionFindReplace.toggled.connect(
            self.code_editor_widget.setSearchBarVisible
        )
        self.code_editor_widget.searchBarToggled.connect(
            self.actionFindReplace.setChecked
        )

        self.actionIncreaseFontSize.triggered.connect(self.editor.zoomIn)
        self.actionDecreaseFontSize.triggered.connect(self.editor.zoomOut)
        self.actionToggleComment.triggered.connect(self.editor.toggleComment)
        self.editor.modificationChanged.connect(self._on_text_modified)

        self.run_dialog = None

        if filePath is not None:
            self._loadFile(filePath)

        self.setHasChanged(False)

    def update_dialog_title(self):
        """
        Updates the script editor dialog title
        """
        if self.code_editor_widget.filePath():
            path, file_name = os.path.split(self.code_editor_widget.filePath())
        else:
            file_name = self.tr("Untitled Script")

        if self.hasChanged:
            file_name = "*" + file_name

        self.setWindowTitle(self.tr("{} - Processing Script Editor").format(file_name))

    def closeEvent(self, event):
        settings = QgsSettings()
        settings.setValue("/Processing/stateScriptEditor", self.saveState())
        settings.setValue("/Processing/geometryScriptEditor", self.saveGeometry())

        if self.hasChanged:
            ret = QMessageBox.question(
                self,
                self.tr("Save Script?"),
                self.tr(
                    "There are unsaved changes in this script. Do you want to keep those?"
                ),
                QMessageBox.StandardButton.Save
                | QMessageBox.StandardButton.Cancel
                | QMessageBox.StandardButton.Discard,
                QMessageBox.StandardButton.Cancel,
            )

            if ret == QMessageBox.StandardButton.Save:
                self.saveScript(False)
                event.accept()
            elif ret == QMessageBox.StandardButton.Discard:
                event.accept()
            else:
                event.ignore()
        else:
            event.accept()

    def openScript(self):
        if self.hasChanged:
            ret = QMessageBox.warning(
                self,
                self.tr("Unsaved changes"),
                self.tr("There are unsaved changes in the script. Continue?"),
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if ret == QMessageBox.StandardButton.No:
                return

        scriptDir = ScriptUtils.scriptsFolders()[0]
        fileName, _ = QFileDialog.getOpenFileName(
            self,
            self.tr("Open script"),
            scriptDir,
            self.tr("Processing scripts (*.py *.PY)"),
        )

        if fileName == "":
            return

        with OverrideCursor(Qt.CursorShape.WaitCursor):
            self._loadFile(fileName)

    def save(self):
        self.saveScript(False)

    def saveAs(self):
        self.saveScript(True)

    def saveScript(self, saveAs):
        newPath = None
        if not self.code_editor_widget.filePath() or saveAs:
            scriptDir = ScriptUtils.scriptsFolders()[0]
            newPath, _ = QFileDialog.getSaveFileName(
                self,
                self.tr("Save script"),
                scriptDir,
                self.tr("Processing scripts (*.py *.PY)"),
            )

            if newPath:
                newPath = QgsFileUtils.ensureFileNameHasExtension(newPath, ["py"])
                self.code_editor_widget.save(newPath)
        elif self.code_editor_widget.filePath():
            self.code_editor_widget.save()

        self.setHasChanged(False)
        QgsApplication.processingRegistry().providerById("script").refreshAlgorithms()

    def _on_text_modified(self, modified):
        self.setHasChanged(modified)

    def setHasChanged(self, hasChanged):
        self.hasChanged = hasChanged
        self.actionSaveScript.setEnabled(hasChanged)
        self.update_dialog_title()

    def runAlgorithm(self):
        if self.run_dialog and not sip.isdeleted(self.run_dialog):
            self.run_dialog.close()
            self.run_dialog = None

        _locals = {}
        try:
            exec(self.editor.text(), _locals)
        except Exception as e:
            error = QgsError(traceback.format_exc(), "Processing")
            QgsErrorDialog.show(error, self.tr("Execution error"))
            return

        alg = None
        try:
            alg = algfactory.instances.pop().createInstance()
        except IndexError:
            for name, attr in _locals.items():
                if (
                    inspect.isclass(attr)
                    and issubclass(
                        attr,
                        (QgsProcessingAlgorithm, QgsProcessingFeatureBasedAlgorithm),
                    )
                    and attr.__name__
                    not in (
                        "QgsProcessingAlgorithm",
                        "QgsProcessingFeatureBasedAlgorithm",
                    )
                ):
                    alg = attr()
                    break

        if alg is None:
            QMessageBox.warning(
                self,
                self.tr("No script found"),
                self.tr("Seems there is no valid script in the file."),
            )
            return

        alg.setProvider(QgsApplication.processingRegistry().providerById("script"))
        alg.initAlgorithm()

        self.run_dialog = alg.createCustomParametersWidget(self)
        if not self.run_dialog:
            self.run_dialog = AlgorithmDialog(alg, parent=self)

        canvas = iface.mapCanvas()
        prevMapTool = canvas.mapTool()

        self.run_dialog.show()

        if canvas.mapTool() != prevMapTool:
            try:
                canvas.mapTool().reset()
            except:
                pass
            canvas.setMapTool(prevMapTool)

    def _loadFile(self, filePath):

        self.code_editor_widget.loadFile(filePath)
        self.hasChanged = False

        self.update_dialog_title()



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""


class HtmlContent:
    def __init__(self, data):
        self.data = data if not isinstance(data, HtmlContent) else data.data

    def toHtml(self):
        if isinstance(self.data, list) or isinstance(self.data, tuple):
            html = ""
            for item in self.data:
                html += HtmlContent(item).toHtml()
            return html

        if hasattr(self.data, "toHtml"):
            return self.data.toHtml()

        html = str(self.data).replace("\n", "<br>")
        return html

    def hasContents(self):
        if isinstance(self.data, list) or isinstance(self.data, tuple):
            empty = True
            for item in self.data:
                if item.hasContents():
                    empty = False
                    break
            return not empty

        if hasattr(self.data, "hasContents"):
            return self.data.hasContents()

        return len(self.data) > 0


class HtmlElem:
    def __init__(self, tag, data, attrs=None):
        self.tag = tag
        self.data = data if isinstance(data, HtmlContent) else HtmlContent(data)
        self.attrs = attrs if attrs is not None else dict()
        if "tag" in self.attrs:
            self.setTag(self.attrs["tag"])
            del self.attrs["tag"]

    def setTag(self, tag):
        self.tag = tag

    def getOriginalData(self):
        return self.data.data

    def setAttr(self, name, value):
        self.attrs[name] = value

    def getAttrsHtml(self):
        html = ""
        for k, v in self.attrs.items():
            html += f' {k}="{v}"'
        return html

    def openTagHtml(self):
        return f"<{self.tag}{self.getAttrsHtml()}>"

    def closeTagHtml(self):
        return "</%s>" % self.tag

    def toHtml(self):
        return f"{self.openTagHtml()}{self.data.toHtml()}{self.closeTagHtml()}"

    def hasContents(self):
        return self.data.toHtml() != ""


class HtmlParagraph(HtmlElem):
    def __init__(self, data, attrs=None):
        HtmlElem.__init__(self, "p", data, attrs)


class HtmlListItem(HtmlElem):
    def __init__(self, data, attrs=None):
        HtmlElem.__init__(self, "li", data, attrs)


class HtmlList(HtmlElem):
    def __init__(self, items, attrs=None):
        # make sure to have HtmlListItem items
        items = list(items)
        for i, item in enumerate(items):
            if not isinstance(item, HtmlListItem):
                items[i] = HtmlListItem(item)
        HtmlElem.__init__(self, "ul", items, attrs)


class HtmlTableCol(HtmlElem):
    def __init__(self, data, attrs=None):
        HtmlElem.__init__(self, "td", data, attrs)

    def closeTagHtml(self):
        # FIX INVALID BEHAVIOR: an empty cell as last table's cell break margins
        return "&nbsp;%s" % HtmlElem.closeTagHtml(self)


class HtmlTableRow(HtmlElem):
    def __init__(self, cols, attrs=None):
        # make sure to have HtmlTableCol items
        cols = list(cols)
        for i, c in enumerate(cols):
            if not isinstance(c, HtmlTableCol):
                cols[i] = HtmlTableCol(c)
        HtmlElem.__init__(self, "tr", cols, attrs)


class HtmlTableHeader(HtmlTableRow):
    def __init__(self, cols, attrs=None):
        HtmlTableRow.__init__(self, cols, attrs)
        for c in self.getOriginalData():
            c.setTag("th")


class HtmlTable(HtmlElem):
    def __init__(self, rows, attrs=None):
        # make sure to have HtmlTableRow items
        rows = list(rows)
        for i, r in enumerate(rows):
            if not isinstance(r, HtmlTableRow):
                rows[i] = HtmlTableRow(r)
        HtmlElem.__init__(self, "table", rows, attrs)


class HtmlSection(HtmlContent):
    def __init__(self, title, content=None):
        data = ['<div class="section"><h2>', title, "</h2>"]
        if content is not None:
            data.extend(["<div>", content, "</div>"])
        data.append("</div>")
        HtmlContent.__init__(self, data)



"""
***************************************************************************
    g_extension_list.py
    ------------------
    Date                 : May 2023
    Copyright            : (C) 2023 by Alister Hood
    Email                : alister.hood at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alister Hood"
__date__ = "May 2023"
__copyright__ = "(C) 2023, Alister Hood"

from qgis.core import QgsProcessingParameterBoolean


def processInputs(alg, parameters, context, feedback):
    # get the option we want to run
    # we would need to update the if statements below if the order in the description file is changed
    selectedOption = alg.parameterAsInt(parameters, "list", context)
    # Locally installed extensions
    if selectedOption == 0:
        optionString = "-a"
    # Extensions available in the official GRASS GIS Addons repository
    elif selectedOption == 1:
        optionString = "-l"
    # Extensions available in the official GRASS GIS Addons repository including module description
    else:
        optionString = "-c"
    param = QgsProcessingParameterBoolean(optionString, "", True)
    alg.addParameter(param)
    # Don't forget to remove the old input parameter
    alg.removeParameter("list")


def convertToHtml(alg, fileName, outputs):
    # don't copy this for something that will have lots of data like r.stats
    # it will be very slow
    with open(fileName) as f:
        outputs["GRASS_ADDONS"] = f.read()
    # this will read the file a second time, but calling it saves us duplicating the code here
    alg.convertToHtml(fileName)



"""
***************************************************************************
    SetRasterStyle.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import (
    QgsProcessingAlgorithm,
    QgsProcessingOutputRasterLayer,
    QgsProcessingParameterFile,
    QgsProcessingParameterRasterLayer,
)
from qgis.PyQt.QtXml import QDomDocument

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class SetRasterStyle(QgisAlgorithm):
    INPUT = "INPUT"
    STYLE = "STYLE"
    OUTPUT = "OUTPUT"

    def group(self):
        return self.tr("Raster tools")

    def groupId(self):
        return "rastertools"

    def __init__(self):
        super().__init__()

    def flags(self):
        return (
            super().flags()
            | QgsProcessingAlgorithm.Flag.FlagNoThreading
            | QgsProcessingAlgorithm.Flag.FlagDeprecated
            | QgsProcessingAlgorithm.Flag.FlagNotAvailableInStandaloneTool
        )

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Raster layer"))
        )
        self.addParameter(
            QgsProcessingParameterFile(
                self.STYLE, self.tr("Style file"), extension="qml"
            )
        )
        self.addOutput(QgsProcessingOutputRasterLayer(self.INPUT, self.tr("Styled")))

    def name(self):
        return "setstyleforrasterlayer"

    def displayName(self):
        return self.tr("Set style for raster layer")

    def shortDescription(self):
        return self.tr("Applies the style from a QML file to a raster layer.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm sets the style of a raster layer. The style must be defined in a QML file."
        )

    def processAlgorithm(self, parameters, context, feedback):
        layer = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        style = self.parameterAsFile(parameters, self.STYLE, context)
        with open(style) as f:
            xml = "".join(f.readlines())
        d = QDomDocument()
        d.setContent(xml)
        layer.importNamedStyle(d)
        layer.triggerRepaint()
        return {self.INPUT: layer}



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : Oct 13, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

The content of this file is based on
- PG_Manager by Martin Dobias (GPLv2 license)
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.PyQt import uic
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtWidgets import QApplication, QDialog, QMessageBox
from qgis.utils import OverrideCursor

from .db_plugins.plugin import DbError
from .dlg_db_error import DlgDbError
from .gui_utils import GuiUtils

Ui_Dialog, _ = uic.loadUiType(GuiUtils.get_ui_file_path("DlgAddGeometryColumn.ui"))


class DlgAddGeometryColumn(QDialog, Ui_Dialog):
    GEOM_TYPES = [
        "POINT",
        "LINESTRING",
        "POLYGON",
        "MULTIPOINT",
        "MULTILINESTRING",
        "MULTIPOLYGON",
        "GEOMETRYCOLLECTION",
    ]

    def __init__(self, parent=None, table=None, db=None):
        QDialog.__init__(self, parent)
        self.table = table
        self.db = self.table.database() if self.table and self.table.database() else db
        self.setupUi(self)

        self.buttonBox.accepted.connect(self.createGeomColumn)

    def createGeomColumn(self):
        """first check whether everything's fine"""
        if self.editName.text() == "":
            QMessageBox.critical(
                self, self.tr("DB Manager"), self.tr("Field name must not be empty.")
            )
            return

        name = self.editName.text()
        geom_type = self.GEOM_TYPES[self.cboType.currentIndex()]
        dim = self.spinDim.value()
        try:
            srid = int(self.editSrid.text())
        except ValueError:
            srid = -1
        createSpatialIndex = False

        # now create the geometry column
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            try:
                self.table.addGeometryColumn(
                    name, geom_type, srid, dim, createSpatialIndex
                )
            except DbError as e:
                DlgDbError.showError(e, self)
                return

        self.accept()



"""
***************************************************************************
    BatchAlgorithmDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import time

from qgis.core import (
    QgsProcessingOutputBoolean,
    QgsProcessingOutputHtml,
    QgsProcessingOutputNumber,
    QgsProcessingOutputString,
    QgsProject,
)
from qgis.gui import QgsProcessingBatchAlgorithmDialogBase
from qgis.utils import iface

from processing.core.ProcessingResults import resultsList
from processing.gui.BatchPanel import BatchPanel
from processing.gui.Postprocessing import handleAlgorithmResults
from processing.tools import dataobjects
from processing.tools.system import getTempFilename


class BatchAlgorithmDialog(QgsProcessingBatchAlgorithmDialogBase):
    def __init__(self, alg, parent=None):
        super().__init__(parent)

        self.setAlgorithm(alg)

        self.setWindowTitle(
            self.tr("Batch Processing - {0}").format(self.algorithm().displayName())
        )
        self.setMainWidget(BatchPanel(self, self.algorithm()))

        self.context = None
        self.hideShortHelp()

    def runAsSingle(self):
        self.close()

        alg_instance = self.algorithm().create()
        dlg = alg_instance.createCustomParametersWidget(parent=iface.mainWindow())
        if not dlg:
            from processing.gui.AlgorithmDialog import AlgorithmDialog

            dlg = AlgorithmDialog(alg_instance, parent=iface.mainWindow())

        dlg.show()
        dlg.exec()

    def processingContext(self):
        if self.context is None:
            self.feedback = self.createFeedback()
            self.context = dataobjects.createContext(self.feedback)
            self.context.setLogLevel(self.logLevel())
        return self.context

    def createContext(self, feedback):
        return dataobjects.createContext(feedback)

    def runAlgorithm(self):
        alg_parameters = []

        load_layers = self.mainWidget().checkLoadLayersOnCompletion.isChecked()
        project = QgsProject.instance() if load_layers else None

        for row in range(self.mainWidget().batchRowCount()):
            parameters, ok = self.mainWidget().parametersForRow(
                row=row,
                context=self.processingContext(),
                destinationProject=project,
                warnOnInvalid=True,
            )
            if ok:
                alg_parameters.append(parameters)
        if not alg_parameters:
            return

        self.execute(alg_parameters)

    def handleAlgorithmResults(self, algorithm, context, feedback, parameters):
        handleAlgorithmResults(algorithm, context, feedback, parameters)

    def loadHtmlResults(self, results, num):
        for out in self.algorithm().outputDefinitions():
            if (
                isinstance(out, QgsProcessingOutputHtml)
                and out.name() in results
                and results[out.name()]
            ):
                resultsList.addResult(
                    icon=self.algorithm().icon(),
                    name=f"{out.description()} [{num}]",
                    result=results[out.name()],
                )

    def createSummaryTable(self, algorithm_results, errors):
        createTable = False

        for out in self.algorithm().outputDefinitions():
            if isinstance(
                out,
                (
                    QgsProcessingOutputNumber,
                    QgsProcessingOutputString,
                    QgsProcessingOutputBoolean,
                ),
            ):
                createTable = True
                break

        if not createTable and not errors:
            return

        outputFile = getTempFilename("html")
        with open(outputFile, "w", encoding="utf-8") as f:
            if createTable:
                for i, res in enumerate(algorithm_results):
                    results = res["results"]
                    params = res["parameters"]
                    if i > 0:
                        f.write("<hr>\n")
                    f.write(self.tr("<h3>Parameters</h3>\n"))
                    f.write("<table>\n")
                    for param in self.algorithm().parameterDefinitions():
                        if not param.isDestination():
                            if param.name() in params:
                                f.write(
                                    f"<tr><th>{param.description()}</th><td>{params[param.name()]}</td></tr>\n"
                                )
                    f.write("</table>\n")
                    f.write(self.tr("<h3>Results</h3>\n"))
                    f.write("<table>\n")
                    for out in self.algorithm().outputDefinitions():
                        if out.name() in results:
                            f.write(
                                f"<tr><th>{out.description()}</th><td>{results[out.name()]}</td></tr>\n"
                            )
                    f.write("</table>\n")
            if errors:
                f.write('<h2 style="color: red">{}</h2>\n'.format(self.tr("Errors")))
            for i, res in enumerate(errors):
                errors = res["errors"]
                params = res["parameters"]
                if i > 0:
                    f.write("<hr>\n")
                f.write(self.tr("<h3>Parameters</h3>\n"))
                f.write("<table>\n")
                for param in self.algorithm().parameterDefinitions():
                    if not param.isDestination():
                        if param.name() in params:
                            f.write(
                                f"<tr><th>{param.description()}</th><td>{params[param.name()]}</td></tr>\n"
                            )
                f.write("</table>\n")
                f.write("<h3>{}</h3>\n".format(self.tr("Error")))
                f.write('<p style="color: red">{}</p>\n'.format("<br>".join(errors)))

        resultsList.addResult(
            icon=self.algorithm().icon(),
            name=f"{self.algorithm().name()} [summary]",
            timestamp=time.localtime(),
            result=outputFile,
        )



"""
***************************************************************************
    AlgorithmDialogBase.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"


class AlgorithmDialogBase:
    class InvalidParameterValue(Exception):
        def __init__(self, param, widget):
            (self.parameter, self.widget) = (param, widget)

    class InvalidOutputExtension(Exception):
        def __init__(self, widget, message):
            self.widget = widget
            self.message = message



"""
***************************************************************************
    ProviderActions.py
    -------------------
    Date                 : April 2017
    Copyright            : (C) 2017 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "April 2017"
__copyright__ = "(C) 2017, Nyall Dason"


class ProviderActions:
    actions = {}

    @staticmethod
    def registerProviderActions(provider, actions):
        """Adds actions for a provider"""
        ProviderActions.actions[provider.id()] = actions

    @staticmethod
    def deregisterProviderActions(provider):
        """Removes actions for a provider"""
        if provider.id() in ProviderActions.actions:
            del ProviderActions.actions[provider.id()]


class ProviderContextMenuActions:
    # All the registered context menu actions for the toolbox
    actions = []

    @staticmethod
    def registerProviderContextMenuActions(actions):
        """Adds context menu actions for a provider"""
        ProviderContextMenuActions.actions.extend(actions)

    @staticmethod
    def deregisterProviderContextMenuActions(actions):
        """Removes context menu actions for a provider"""
        for act in actions:
            ProviderContextMenuActions.actions.remove(act)



"""
***************************************************************************
    ModelerParameterDefinitionDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import math

from qgis.core import (
    QgsApplication,
    QgsProcessing,
    QgsProcessingDestinationParameter,
    QgsProcessingModelAlgorithm,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterFolderDestination,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterVectorDestination,
    QgsSettings,
)
from qgis.gui import (
    QgsColorButton,
    QgsGui,
    QgsProcessingLayerOutputDestinationWidget,
    QgsProcessingModelConfigWidget,
)
from qgis.PyQt.QtCore import QByteArray, QCoreApplication, Qt
from qgis.PyQt.QtGui import QColor
from qgis.PyQt.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from processing.core import parameters
from processing.modeler.exceptions import UndefinedParameterException


class ModelerParameterDefinitionWidget(QgsProcessingModelConfigWidget):
    def __init__(self, alg, paramType=None, param=None):
        self.alg = alg
        self.paramType = paramType
        self.param = param.clone() if param else None
        QgsProcessingModelConfigWidget.__init__(self)
        self.setupUi(self.param, self.paramType)

    def switchToCommentTab(self):
        self.tab.setCurrentIndex(1)
        self.commentEdit.setFocus()
        self.commentEdit.selectAll()

    def setupUi(self, existing_param, paramType):
        type_metadata = QgsApplication.processingRegistry().parameterType(
            existing_param.type() if existing_param else paramType
        )
        self.setPanelTitle(
            self.tr("{} Parameter Definition").format(type_metadata.name())
        )

        self.mainLayout = QVBoxLayout()
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.tab = QTabWidget()
        self.mainLayout.addWidget(self.tab)

        self.verticalLayout = QVBoxLayout()

        self.label = QLabel(self.tr("Parameter name"))
        self.verticalLayout.addWidget(self.label)
        self.nameTextBox = QLineEdit()
        self.verticalLayout.addWidget(self.nameTextBox)

        if isinstance(existing_param, QgsProcessingParameterDefinition):
            self.nameTextBox.setText(existing_param.description())

        self.nameTextBox.textChanged.connect(self.widgetChanged)

        if isinstance(existing_param, QgsProcessingDestinationParameter):
            self.verticalLayout.addWidget(QLabel(self.tr("Default value")))
            self.defaultWidget = QgsProcessingLayerOutputDestinationWidget(
                existing_param, defaultSelection=True
            )
            self.verticalLayout.addWidget(self.defaultWidget)
            # TODO check
            self.defaultWidget.destinationChanged.connect(self.widgetChanged)

        self.verticalLayout.addSpacing(20)
        self.requiredCheck = QCheckBox()
        self.requiredCheck.setText(self.tr("Mandatory"))
        self.requiredCheck.setChecked(True)
        if existing_param is not None:
            self.requiredCheck.setChecked(
                not existing_param.flags()
                & QgsProcessingParameterDefinition.Flag.FlagOptional
            )
        self.verticalLayout.addWidget(self.requiredCheck)
        self.requiredCheck.toggled.connect(self.widgetChanged)

        self.advancedCheck = QCheckBox()
        self.advancedCheck.setText(self.tr("Advanced"))
        self.advancedCheck.setChecked(False)
        if existing_param is not None:
            self.advancedCheck.setChecked(
                existing_param.flags()
                & QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
        self.verticalLayout.addWidget(self.advancedCheck)
        self.advancedCheck.toggled.connect(self.widgetChanged)

        # If child algorithm output is mandatory, disable checkbox
        if isinstance(existing_param, QgsProcessingDestinationParameter):
            child = self.alg.childAlgorithms()[
                existing_param.metadata()["_modelChildId"]
            ]
            model_output = child.modelOutput(
                existing_param.metadata()["_modelChildOutputName"]
            )
            param_def = child.algorithm().parameterDefinition(
                model_output.childOutputName()
            )
            if not (
                param_def.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
            ):
                self.requiredCheck.setEnabled(False)
                self.requiredCheck.setChecked(True)

            self.advancedCheck.setEnabled(False)
            self.advancedCheck.setChecked(False)

        self.verticalLayout.addStretch()

        w = QWidget()
        w.setLayout(self.verticalLayout)
        self.tab.addTab(w, self.tr("Properties"))

        self.commentLayout = QVBoxLayout()
        self.commentEdit = QTextEdit()
        self.commentEdit.setAcceptRichText(False)
        self.commentLayout.addWidget(self.commentEdit, 1)
        self.commentEdit.textChanged.connect(self.widgetChanged)

        hl = QHBoxLayout()
        hl.setContentsMargins(0, 0, 0, 0)
        hl.addWidget(QLabel(self.tr("Color")))
        self.comment_color_button = QgsColorButton()
        self.comment_color_button.setAllowOpacity(True)
        self.comment_color_button.setWindowTitle(self.tr("Comment Color"))
        self.comment_color_button.setShowNull(True, self.tr("Default"))
        hl.addWidget(self.comment_color_button)
        self.commentLayout.addLayout(hl)
        self.comment_color_button.colorChanged.connect(self.widgetChanged)

        w2 = QWidget()
        w2.setLayout(self.commentLayout)
        self.tab.addTab(w2, self.tr("Comments"))

        self.setLayout(self.mainLayout)

    def setComments(self, text):
        self.commentEdit.setPlainText(text)

    def comments(self):
        return self.commentEdit.toPlainText()

    def setCommentColor(self, color):
        if color.isValid():
            self.comment_color_button.setColor(color)
        else:
            self.comment_color_button.setToNull()

    def commentColor(self):
        return (
            self.comment_color_button.color()
            if not self.comment_color_button.isNull()
            else QColor()
        )

    def create_parameter(self):
        description = self.nameTextBox.text()

        safeName = QgsProcessingModelAlgorithm.safeName(description)
        name = safeName.lower()

        # Destination parameter
        new_param = None
        if isinstance(self.param, QgsProcessingParameterFeatureSink):
            new_param = QgsProcessingParameterFeatureSink(
                name=name,
                description=description,
                type=self.param.dataType(),
                defaultValue=self.defaultWidget.value(),
            )
        elif isinstance(self.param, QgsProcessingParameterFileDestination):
            new_param = QgsProcessingParameterFileDestination(
                name=name,
                description=description,
                fileFilter=self.param.fileFilter(),
                defaultValue=self.defaultWidget.value(),
            )
        elif isinstance(self.param, QgsProcessingParameterFolderDestination):
            new_param = QgsProcessingParameterFolderDestination(
                name=name,
                description=description,
                defaultValue=self.defaultWidget.value(),
            )
        elif isinstance(self.param, QgsProcessingParameterRasterDestination):
            new_param = QgsProcessingParameterRasterDestination(
                name=name,
                description=description,
                defaultValue=self.defaultWidget.value(),
            )
        elif isinstance(self.param, QgsProcessingParameterVectorDestination):
            new_param = QgsProcessingParameterVectorDestination(
                name=name,
                description=description,
                type=self.param.dataType(),
                defaultValue=self.defaultWidget.value(),
            )

        else:
            if self.paramType:
                typeId = self.paramType
            else:
                typeId = self.param.type()

            paramTypeDef = (
                QgsApplication.instance().processingRegistry().parameterType(typeId)
            )
            if not paramTypeDef:
                msg = self.tr(
                    "The parameter `{}` is not registered, are you missing a required plugin?"
                ).format(typeId)
                raise UndefinedParameterException(msg)
            new_param = paramTypeDef.create(name)
            new_param.setDescription(description)
            new_param.setMetadata(paramTypeDef.metadata())

        if not self.requiredCheck.isChecked():
            new_param.setFlags(
                new_param.flags() | QgsProcessingParameterDefinition.Flag.FlagOptional
            )
        else:
            new_param.setFlags(
                new_param.flags() & ~QgsProcessingParameterDefinition.Flag.FlagOptional
            )

        if self.advancedCheck.isChecked():
            new_param.setFlags(
                new_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
        else:
            new_param.setFlags(
                new_param.flags() & ~QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )

        return new_param


class ModelerParameterDefinitionDialog(QDialog):
    @staticmethod
    def use_legacy_dialog(param=None, paramType=None):
        if isinstance(param, QgsProcessingDestinationParameter):
            return True

        # yay, use new API!
        return False

    def __init__(self, alg, paramType=None, param=None):
        QDialog.__init__(self)

        self.widget = ModelerParameterDefinitionWidget(alg, paramType, param)
        self.setObjectName("ModelerParameterDefinitionDialog")
        self.setModal(True)
        self.setupUi()

        QgsGui.enableAutoGeometryRestore(self)

    def switchToCommentTab(self):
        self.widget.switchToCommentTab()

    def setupUi(self):
        self.mainLayout = QVBoxLayout()
        self.mainLayout.addWidget(self.widget, 1)

        self.setMinimumWidth(300)

        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(
            QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok
        )
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.mainLayout.addWidget(self.buttonBox)

        self.setLayout(self.mainLayout)

        self.widget.panelAccepted.connect(self.reject)

    def setComments(self, text):
        self.widget.setComments(text)

    def comments(self):
        return self.widget.comments()

    def setCommentColor(self, color):
        self.widget.setCommentColor(color)

    def commentColor(self):
        return self.widget.commentColor()

    def create_parameter(self):
        return self.widget.create_parameter()

    def accept(self):
        description = self.widget.nameTextBox.text()
        if description.strip() == "":
            QMessageBox.warning(
                self,
                self.tr("Unable to define parameter"),
                self.tr("Invalid parameter name"),
            )
            return

        super().accept()



###############################################################################
#
# CSW Client
# ---------------------------------------------------------
# QGIS Catalog Service client.
#
# Copyright (C) 2010 NextGIS (http://nextgis.org),
#                    Alexander Bruy (alexander.bruy@gmail.com),
#                    Maxim Dubinin (sim@gis-lab.info)
#
# Copyright (C) 2017 Tom Kralidis (tomkralidis@gmail.com)
#
# This source is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This code is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
###############################################################################

import json
import os.path
from urllib.request import ProxyHandler, build_opener, install_opener

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsCoordinateReferenceSystem,
    QgsCoordinateTransform,
    QgsGeometry,
    QgsPointXY,
    QgsProject,
    QgsProviderRegistry,
    QgsRectangle,
    QgsSettings,
    QgsSettingsTree,
)
from qgis.gui import QgsGui, QgsRubberBand
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QColor
from qgis.PyQt.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QMessageBox,
    QTreeWidgetItem,
    QWidget,
)
from qgis.utils import OverrideCursor

try:
    from owslib.util import Authentication
except ImportError:
    pass

from MetaSearch import link_types
from MetaSearch.dialogs.apidialog import APIRequestResponseDialog
from MetaSearch.dialogs.manageconnectionsdialog import ManageConnectionsDialog
from MetaSearch.dialogs.newconnectiondialog import NewConnectionDialog
from MetaSearch.dialogs.recorddialog import RecordDialog
from MetaSearch.search_backend import get_catalog_service
from MetaSearch.util import (
    StaticContext,
    clean_ows_url,
    get_connections_from_file,
    get_help_url,
    get_ui_class,
    normalize_text,
    open_url,
    render_template,
    serialize_string,
)

BASE_CLASS = get_ui_class("maindialog.ui")


class MetaSearchDialog(QDialog, BASE_CLASS):
    """main dialogue"""

    def __init__(self, iface):
        """init window"""

        QDialog.__init__(self)
        self.setupUi(self)

        self.iface = iface
        self.map = iface.mapCanvas()
        self.settings = QgsSettings()
        self.catalog = None
        self.catalog_url = None
        self.catalog_username = None
        self.catalog_password = None
        self.catalog_type = None
        self.context = StaticContext()

        self.leKeywords.setShowSearchIcon(True)
        self.leKeywords.setPlaceholderText(self.tr("Search keywords"))

        self.setWindowTitle(self.tr("MetaSearch"))

        self.rubber_band = QgsRubberBand(self.map, Qgis.GeometryType.Polygon)
        self.rubber_band.setColor(QColor(255, 0, 0, 75))
        self.rubber_band.setWidth(5)

        # form inputs
        self.startfrom = 1
        self.constraints = []
        self.maxrecords = int(self.settings.value("/MetaSearch/returnRecords", 10))
        self.timeout = int(self.settings.value("/MetaSearch/timeout", 10))
        self.disable_ssl_verification = self.settings.value(
            "/MetaSearch/disableSSL", False, bool
        )
        self.log_debugging_messages = self.settings.value(
            "/MetaSearch/logDebugging", False, bool
        )

        # Services tab
        self.cmbConnectionsServices.activated.connect(self.save_connection)
        self.cmbConnectionsSearch.activated.connect(self.save_connection)
        self.btnServerInfo.clicked.connect(self.connection_info)
        self.btnAddDefault.clicked.connect(self.add_default_connections)
        self.btnRawAPIResponse.clicked.connect(self.show_api)
        self.tabWidget.currentChanged.connect(self.populate_connection_list)

        # server management buttons
        self.btnNew.clicked.connect(self.add_connection)
        self.btnEdit.clicked.connect(self.edit_connection)
        self.btnDelete.clicked.connect(self.delete_connection)
        self.btnLoad.clicked.connect(self.load_connections)
        self.btnSave.clicked.connect(save_connections)

        # Search tab
        self.treeRecords.itemSelectionChanged.connect(self.record_clicked)
        self.treeRecords.itemDoubleClicked.connect(self.show_metadata)
        self.btnSearch.clicked.connect(self.search)
        self.leKeywords.returnPressed.connect(self.search)
        # prevent dialog from closing upon pressing enter
        self.buttonBox.button(QDialogButtonBox.StandardButton.Close).setAutoDefault(
            False
        )
        # launch help from button
        self.buttonBox.helpRequested.connect(self.help)
        self.btnCanvasBbox.setAutoDefault(False)
        self.btnCanvasBbox.clicked.connect(self.set_bbox_from_map)
        self.btnGlobalBbox.clicked.connect(self.set_bbox_global)

        # navigation buttons
        self.btnFirst.clicked.connect(self.navigate)
        self.btnPrev.clicked.connect(self.navigate)
        self.btnNext.clicked.connect(self.navigate)
        self.btnLast.clicked.connect(self.navigate)

        self.mActionAddWms.triggered.connect(self.add_to_ows)
        self.mActionAddWfs.triggered.connect(self.add_to_ows)
        self.mActionAddWcs.triggered.connect(self.add_to_ows)
        self.mActionAddAms.triggered.connect(self.add_to_ows)
        self.mActionAddAfs.triggered.connect(self.add_to_ows)
        self.mActionAddGisFile.triggered.connect(self.add_gis_file)
        self.btnViewRawAPIResponse.clicked.connect(self.show_api)

        self.manageGui()

    def manageGui(self):
        """open window"""

        def _on_timeout_change(value):
            self.settings.setValue("/MetaSearch/timeout", value)
            self.timeout = value

        def _on_records_change(value):
            self.settings.setValue("/MetaSearch/returnRecords", value)
            self.maxrecords = value

        def _on_ssl_state_change(state):
            self.settings.setValue("/MetaSearch/disableSSL", bool(state))
            self.disable_ssl_verification = bool(state)

        def _on_debugging_state_change(state):
            self.settings.setValue("/MetaSearch/logDebugging", bool(state))
            self.log_debugging_messages = bool(state)

        self.tabWidget.setCurrentIndex(0)
        self.populate_connection_list()
        self.btnRawAPIResponse.setEnabled(False)

        # load settings
        self.spnRecords.setValue(self.maxrecords)
        self.spnRecords.valueChanged.connect(_on_records_change)
        self.spnTimeout.setValue(self.timeout)
        self.spnTimeout.valueChanged.connect(_on_timeout_change)
        self.disableSSLVerification.setChecked(self.disable_ssl_verification)
        self.disableSSLVerification.stateChanged.connect(_on_ssl_state_change)
        self.logDebuggingMessages.setChecked(self.log_debugging_messages)
        self.logDebuggingMessages.stateChanged.connect(_on_debugging_state_change)

        key = "/MetaSearch/%s" % self.cmbConnectionsSearch.currentText()
        self.catalog_url = self.settings.value("%s/url" % key)
        self.catalog_username = self.settings.value("%s/username" % key)
        self.catalog_password = self.settings.value("%s/password" % key)
        self.catalog_type = self.settings.value("%s/catalog-type" % key)

        self.set_bbox_global()

        self.reset_buttons()

        # install proxy handler if specified in QGIS settings
        self.install_proxy()

    # Services tab

    def populate_connection_list(self):
        """populate select box with connections"""

        self.settings.beginGroup("/MetaSearch/")
        self.cmbConnectionsServices.clear()
        self.cmbConnectionsServices.addItems(self.settings.childGroups())
        self.cmbConnectionsSearch.clear()
        self.cmbConnectionsSearch.addItems(self.settings.childGroups())
        self.settings.endGroup()

        self.set_connection_list_position()

        if self.cmbConnectionsServices.count() == 0:
            # no connections - disable various buttons
            state_disabled = False
            self.btnSave.setEnabled(state_disabled)
            # and start with connection tab open
            self.tabWidget.setCurrentIndex(1)
            # tell the user to add services
            msg = self.tr(
                "No services/connections defined. To get "
                "started with MetaSearch, create a new "
                "connection by clicking 'New' or click "
                "'Add default services'."
            )
            self.textMetadata.setHtml("<p><h3>%s</h3></p>" % msg)
        else:
            # connections - enable various buttons
            state_disabled = True

        self.btnServerInfo.setEnabled(state_disabled)
        self.btnEdit.setEnabled(state_disabled)
        self.btnDelete.setEnabled(state_disabled)

    def set_connection_list_position(self):
        """set the current index to the selected connection"""
        to_select = self.settings.value("/MetaSearch/selected")
        conn_count = self.cmbConnectionsServices.count()

        if conn_count == 0:
            self.btnDelete.setEnabled(False)
            self.btnServerInfo.setEnabled(False)
            self.btnEdit.setEnabled(False)

        # does to_select exist in cmbConnectionsServices?
        exists = False
        for i in range(conn_count):
            if self.cmbConnectionsServices.itemText(i) == to_select:
                self.cmbConnectionsServices.setCurrentIndex(i)
                self.cmbConnectionsSearch.setCurrentIndex(i)
                exists = True
                break

        # If we couldn't find the stored item, but there are some, default
        # to the last item (this makes some sense when deleting items as it
        # allows the user to repeatidly click on delete to remove a whole
        # lot of items)
        if not exists and conn_count > 0:
            # If to_select is null, then the selected connection wasn't found
            # by QgsSettings, which probably means that this is the first time
            # the user has used CSWClient, so default to the first in the list
            # of connetions. Otherwise default to the last.
            if not to_select:
                current_index = 0
            else:
                current_index = conn_count - 1

            self.cmbConnectionsServices.setCurrentIndex(current_index)
            self.cmbConnectionsSearch.setCurrentIndex(current_index)

    def save_connection(self):
        """save connection"""

        caller = self.sender().objectName()

        if caller == "cmbConnectionsServices":  # servers tab
            current_text = self.cmbConnectionsServices.currentText()
        elif caller == "cmbConnectionsSearch":  # search tab
            current_text = self.cmbConnectionsSearch.currentText()

        self.settings.setValue("/MetaSearch/selected", current_text)
        key = "/MetaSearch/%s" % current_text

        if caller == "cmbConnectionsSearch":  # bind to service in search tab
            self.catalog_url = self.settings.value("%s/url" % key)
            self.catalog_username = self.settings.value("%s/username" % key)
            self.catalog_password = self.settings.value("%s/password" % key)
            self.catalog_type = self.settings.value("%s/catalog-type" % key)

        if caller == "cmbConnectionsServices":  # clear server metadata
            self.textMetadata.clear()

        self.btnRawAPIResponse.setEnabled(False)

    def connection_info(self):
        """show connection info"""

        current_text = self.cmbConnectionsServices.currentText()
        key = "/MetaSearch/%s" % current_text
        self.catalog_url = self.settings.value("%s/url" % key)
        self.catalog_username = self.settings.value("%s/username" % key)
        self.catalog_password = self.settings.value("%s/password" % key)
        self.catalog_type = self.settings.value("%s/catalog-type" % key)

        # connect to the server
        if not self._get_catalog():
            return

        if self.catalog:  # display service metadata
            self.btnRawAPIResponse.setEnabled(True)
            metadata = render_template(
                "en",
                self.context,
                self.catalog.conn,
                self.catalog.service_info_template,
            )
            style = QgsApplication.reportStyleSheet()
            self.textMetadata.clear()
            self.textMetadata.document().setDefaultStyleSheet(style)
            self.textMetadata.setHtml(metadata)

            # clear results and disable buttons in Search tab
            self.clear_results()

    def add_connection(self):
        """add new service"""

        conn_new = NewConnectionDialog()
        conn_new.setWindowTitle(self.tr("New Catalog Service"))
        if conn_new.exec() == QDialog.DialogCode.Accepted:  # add to service list
            self.populate_connection_list()
        self.textMetadata.clear()

    def edit_connection(self):
        """modify existing connection"""

        current_text = self.cmbConnectionsServices.currentText()

        url = self.settings.value("/MetaSearch/%s/url" % current_text)

        conn_edit = NewConnectionDialog(current_text)
        conn_edit.setWindowTitle(self.tr("Edit Catalog Service"))
        conn_edit.leName.setText(current_text)
        conn_edit.leURL.setText(url)
        conn_edit.leUsername.setText(
            self.settings.value("/MetaSearch/%s/username" % current_text)
        )
        conn_edit.lePassword.setText(
            self.settings.value("/MetaSearch/%s/password" % current_text)
        )

        conn_edit.cmbCatalogType.setCurrentText(
            self.settings.value("/MetaSearch/%s/catalog-type" % current_text)
        )

        if conn_edit.exec() == QDialog.DialogCode.Accepted:  # update service list
            self.populate_connection_list()

    def delete_connection(self):
        """delete connection"""

        current_text = self.cmbConnectionsServices.currentText()

        key = "/MetaSearch/%s" % current_text

        msg = self.tr("Remove service {0}?").format(current_text)

        result = QMessageBox.question(
            self,
            self.tr("Delete Service"),
            msg,
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )
        if result == QMessageBox.StandardButton.Yes:  # remove service from list
            self.settings.remove(key)
            index_to_delete = self.cmbConnectionsServices.currentIndex()
            self.cmbConnectionsServices.removeItem(index_to_delete)
            self.cmbConnectionsSearch.removeItem(index_to_delete)
            self.set_connection_list_position()

    def load_connections(self):
        """load services from list"""

        ManageConnectionsDialog(1).exec()
        self.populate_connection_list()

    def add_default_connections(self):
        """add default connections"""

        filename = os.path.join(
            self.context.ppath, "resources", "connections-default.xml"
        )

        doc = get_connections_from_file(self, filename)
        if doc is None:
            return

        self.settings.beginGroup("/MetaSearch/")
        keys = self.settings.childGroups()
        self.settings.endGroup()

        for server in doc.findall("csw"):
            name = server.attrib.get("name")
            # check for duplicates
            if name in keys:
                msg = self.tr("{0} exists.  Overwrite?").format(name)
                res = QMessageBox.warning(
                    self,
                    self.tr("Loading connections"),
                    msg,
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                )
                if res != QMessageBox.StandardButton.Yes:
                    continue

            # no dups detected or overwrite is allowed
            key = "/MetaSearch/%s" % name
            self.settings.setValue("%s/url" % key, server.attrib.get("url"))
            self.settings.setValue(
                "%s/catalog-type" % key, server.attrib.get("type", "OGC CSW 2.0.2")
            )

        self.populate_connection_list()

    # Settings tab

    def set_ows_save_title_ask(self):
        """save ows save strategy as save ows title, ask if duplicate"""

        self.settings.setValue("/MetaSearch/ows_save_strategy", "title_ask")

    def set_ows_save_title_no_ask(self):
        """save ows save strategy as save ows title, do NOT ask if duplicate"""

        self.settings.setValue("/MetaSearch/ows_save_strategy", "title_no_ask")

    def set_ows_save_temp_name(self):
        """save ows save strategy as save with a temporary name"""

        self.settings.setValue("/MetaSearch/ows_save_strategy", "temp_name")

    # Search tab

    def set_bbox_from_map(self):
        """set bounding box from map extent"""

        crs = self.map.mapSettings().destinationCrs()
        crsid = crs.authid()

        extent = self.map.extent()

        if crsid != "EPSG:4326":  # reproject to EPSG:4326
            src = QgsCoordinateReferenceSystem(crsid)
            dest = QgsCoordinateReferenceSystem("EPSG:4326")
            xform = QgsCoordinateTransform(src, dest, QgsProject.instance())
            minxy = xform.transform(QgsPointXY(extent.xMinimum(), extent.yMinimum()))
            maxxy = xform.transform(QgsPointXY(extent.xMaximum(), extent.yMaximum()))
            minx, miny = minxy
            maxx, maxy = maxxy
        else:  # EPSG:4326
            minx = extent.xMinimum()
            miny = extent.yMinimum()
            maxx = extent.xMaximum()
            maxy = extent.yMaximum()

        self.leNorth.setText(str(maxy)[0:9])
        self.leSouth.setText(str(miny)[0:9])
        self.leWest.setText(str(minx)[0:9])
        self.leEast.setText(str(maxx)[0:9])

    def set_bbox_global(self):
        """set global bounding box"""
        self.leNorth.setText("90")
        self.leSouth.setText("-90")
        self.leWest.setText("-180")
        self.leEast.setText("180")

    def search(self):
        """execute search"""

        self.catalog = None
        self.constraints = []

        # clear all fields and disable buttons
        self.clear_results()

        # set current catalog
        current_text = self.cmbConnectionsSearch.currentText()
        key = "/MetaSearch/%s" % current_text
        self.catalog_url = self.settings.value("%s/url" % key)
        self.catalog_username = self.settings.value("%s/username" % key)
        self.catalog_password = self.settings.value("%s/password" % key)
        self.catalog_type = self.settings.value("%s/catalog-type" % key)

        # start position and number of records to return
        self.startfrom = 1

        # bbox
        # CRS is WGS84 with axis order longitude, latitude
        # defined by 'urn:ogc:def:crs:OGC:1.3:CRS84'
        minx = self.leWest.text()
        miny = self.leSouth.text()
        maxx = self.leEast.text()
        maxy = self.leNorth.text()
        bbox = [minx, miny, maxx, maxy]
        keywords = self.leKeywords.text()

        # build request
        if not self._get_catalog():
            return

        # TODO: allow users to select resources types
        # to find ('service', 'dataset', etc.)
        try:
            with OverrideCursor(Qt.CursorShape.WaitCursor):
                self.catalog.query_records(
                    bbox, keywords, self.maxrecords, self.startfrom
                )

        except Exception as err:
            QMessageBox.warning(
                self, self.tr("Search error"), self.tr("Search error: {0}").format(err)
            )
            return

        if self.catalog.matches == 0:
            self.lblResults.setText(self.tr("0 results"))
            return

        self.display_results()

    def display_results(self):
        """display search results"""

        self.treeRecords.clear()

        position = self.catalog.returned + self.startfrom - 1

        msg = self.tr(
            "Showing {0} - {1} of %n result(s)",
            "number of results",
            self.catalog.matches,
        ).format(self.startfrom, position)

        self.lblResults.setText(msg)

        for rec in self.catalog.records():
            item = QTreeWidgetItem(self.treeRecords)
            if rec["type"]:
                item.setText(0, normalize_text(rec["type"]))
            else:
                item.setText(0, "unknown")
            if rec["title"]:
                item.setText(1, normalize_text(rec["title"]))
            if rec["identifier"]:
                set_item_data(item, "identifier", rec["identifier"])

        self.btnViewRawAPIResponse.setEnabled(True)

        if self.catalog.matches < self.maxrecords:
            disabled = False
        else:
            disabled = True

        self.btnFirst.setEnabled(disabled)
        self.btnPrev.setEnabled(disabled)
        self.btnNext.setEnabled(disabled)
        self.btnLast.setEnabled(disabled)
        self.btnRawAPIResponse.setEnabled(False)

    def clear_results(self):
        """clear search results"""

        self.lblResults.clear()
        self.treeRecords.clear()
        self.reset_buttons()

    def record_clicked(self):
        """record clicked signal"""

        # disable only service buttons
        self.reset_buttons(True, False, False)

        self.rubber_band.reset()

        if not self.treeRecords.selectedItems():
            return

        item = self.treeRecords.currentItem()
        if not item:
            return

        identifier = get_item_data(item, "identifier")
        try:
            record = next(
                item
                for item in self.catalog.records()
                if item["identifier"] == identifier
            )
        except KeyError:
            QMessageBox.warning(
                self,
                self.tr("Record parsing error"),
                "Unable to locate record identifier",
            )
            return

        # if the record has a bbox, show a footprint on the map
        if record["bbox"] is not None:
            bx = record["bbox"]
            rt = QgsRectangle(
                float(bx["minx"]),
                float(bx["miny"]),
                float(bx["maxx"]),
                float(bx["maxy"]),
            )
            geom = QgsGeometry.fromRect(rt)

            if geom is not None:
                src = QgsCoordinateReferenceSystem("EPSG:4326")
                dst = self.map.mapSettings().destinationCrs()
                if src.postgisSrid() != dst.postgisSrid():
                    ctr = QgsCoordinateTransform(src, dst, QgsProject.instance())
                    try:
                        geom.transform(ctr)
                    except Exception as err:
                        QMessageBox.warning(
                            self, self.tr("Coordinate Transformation Error"), str(err)
                        )
                self.rubber_band.setToGeometry(geom, None)

        # figure out if the data is interactive and can be operated on
        self.find_services(record, item)

    def find_services(self, record, item):
        """scan record for WMS/WMTS|WFS|WCS endpoints"""

        services = {}
        for link in record["links"]:
            link = self.catalog.parse_link(link)
            if "scheme" in link:
                link_type = link["scheme"]
            elif "protocol" in link:
                link_type = link["protocol"]
            else:
                link_type = None

            if link_type is not None:
                link_type = link_type.upper()

            wmswmst_link_types = list(map(str.upper, link_types.WMSWMST_LINK_TYPES))
            wfs_link_types = list(map(str.upper, link_types.WFS_LINK_TYPES))
            wcs_link_types = list(map(str.upper, link_types.WCS_LINK_TYPES))
            ams_link_types = list(map(str.upper, link_types.AMS_LINK_TYPES))
            afs_link_types = list(map(str.upper, link_types.AFS_LINK_TYPES))
            gis_file_link_types = list(map(str.upper, link_types.GIS_FILE_LINK_TYPES))

            # if the link type exists, and it is one of the acceptable
            # interactive link types, then set
            all_link_types = (
                wmswmst_link_types
                + wfs_link_types
                + wcs_link_types
                + ams_link_types
                + afs_link_types
                + gis_file_link_types
            )

            if all([link_type is not None, link_type in all_link_types]):
                if link_type in wmswmst_link_types:
                    services["wms"] = link["url"]
                    self.mActionAddWms.setEnabled(True)
                if link_type in wfs_link_types:
                    services["wfs"] = link["url"]
                    self.mActionAddWfs.setEnabled(True)
                if link_type in wcs_link_types:
                    services["wcs"] = link["url"]
                    self.mActionAddWcs.setEnabled(True)
                if link_type in ams_link_types:
                    services["ams"] = link["url"]
                    self.mActionAddAms.setEnabled(True)
                if link_type in afs_link_types:
                    services["afs"] = link["url"]
                    self.mActionAddAfs.setEnabled(True)
                if link_type in gis_file_link_types:
                    services["gis_file"] = link["url"]
                    services["title"] = record.get("title", "")
                    self.mActionAddGisFile.setEnabled(True)
                self.tbAddData.setEnabled(True)

            set_item_data(item, "link", json.dumps(services))

    def navigate(self):
        """manage navigation / paging"""

        caller = self.sender().objectName()

        if caller == "btnFirst":
            self.startfrom = 1
        elif caller == "btnLast":
            self.startfrom = self.catalog.matches - self.maxrecords + 1
        elif caller == "btnNext":
            if self.startfrom > self.catalog.matches - self.maxrecords:
                msg = self.tr("End of results. Go to start?")
                res = QMessageBox.information(
                    self,
                    self.tr("Navigation"),
                    msg,
                    (QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel),
                )
                if res == QMessageBox.StandardButton.Ok:
                    self.startfrom = 1
                else:
                    return
            else:
                self.startfrom += self.maxrecords
        elif caller == "btnPrev":
            if self.startfrom == 1:
                msg = self.tr("Start of results. Go to end?")
                res = QMessageBox.information(
                    self,
                    self.tr("Navigation"),
                    msg,
                    (QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel),
                )
                if res == QMessageBox.StandardButton.Ok:
                    self.startfrom = self.catalog.matches - self.maxrecords + 1
                else:
                    return
            elif self.startfrom <= self.maxrecords:
                self.startfrom = 1
            else:
                self.startfrom -= self.maxrecords

        # bbox
        # CRS is WGS84 with axis order longitude, latitude
        # defined by 'urn:ogc:def:crs:OGC:1.3:CRS84'
        minx = self.leWest.text()
        miny = self.leSouth.text()
        maxx = self.leEast.text()
        maxy = self.leNorth.text()
        bbox = [minx, miny, maxx, maxy]
        keywords = self.leKeywords.text()

        try:
            with OverrideCursor(Qt.CursorShape.WaitCursor):
                self.catalog.query_records(
                    bbox, keywords, limit=self.maxrecords, offset=self.startfrom
                )
        except Exception as err:
            QMessageBox.warning(
                self, self.tr("Search error"), self.tr("Search error: {0}").format(err)
            )
            return

        self.display_results()

    def add_to_ows(self):
        """add to OWS provider connection list"""

        conn_name_matches = []

        item = self.treeRecords.currentItem()

        if not item:
            return

        item_data = json.loads(get_item_data(item, "link"))

        caller = self.sender().objectName()

        if caller == "mActionAddWms":
            service_type = "OGC:WMS/OGC:WMTS"
            sname = "WMS"
            dyn_param = ["wms"]
            provider_name = "wms"
            setting_node = (
                QgsSettingsTree.node("connections")
                .childNode("ows")
                .childNode("connections")
            )
            data_url = item_data["wms"]
        elif caller == "mActionAddWfs":
            service_type = "OGC:WFS"
            sname = "WFS"
            dyn_param = ["wfs"]
            provider_name = "WFS"
            setting_node = (
                QgsSettingsTree.node("connections")
                .childNode("ows")
                .childNode("connections")
            )
            data_url = item_data["wfs"]
        elif caller == "mActionAddWcs":
            service_type = "OGC:WCS"
            sname = "WCS"
            dyn_param = ["wcs"]
            provider_name = "wcs"
            setting_node = (
                QgsSettingsTree.node("connections")
                .childNode("ows")
                .childNode("connections")
            )
            data_url = item_data["wcs"]
        elif caller == "mActionAddAfs":
            service_type = "ESRI:ArcGIS:FeatureServer"
            sname = "AFS"
            dyn_param = []
            provider_name = "arcgisfeatureserver"
            setting_node = QgsSettingsTree.node("connections").childNode(
                "arcgisfeatureserver"
            )
            data_url = item_data["afs"].split("FeatureServer")[0] + "FeatureServer"

        keys = setting_node.items(dyn_param)

        sname = "%s from MetaSearch" % sname
        for key in keys:
            if key.startswith(sname):
                conn_name_matches.append(key)
        if conn_name_matches:
            sname = conn_name_matches[-1]

        # check for duplicates
        if sname in keys:  # duplicate found
            msg = self.tr("Connection {0} exists. Overwrite?").format(sname)
            res = QMessageBox.warning(
                self,
                self.tr("Saving server"),
                msg,
                QMessageBox.StandardButton.Yes
                | QMessageBox.StandardButton.No
                | QMessageBox.StandardButton.Cancel,
            )
            if res == QMessageBox.StandardButton.No:  # assign new name with serial
                sname = serialize_string(sname)
            elif res == QMessageBox.StandardButton.Cancel:
                return

        # no dups detected or overwrite is allowed
        dyn_param.append(sname)
        setting_node.childSetting("url").setValue(clean_ows_url(data_url), dyn_param)

        # open provider window
        ows_provider = QgsGui.sourceSelectProviderRegistry().createSelectionWidget(
            provider_name,
            self,
            Qt.WindowType.Widget,
            QgsProviderRegistry.WidgetMode.Embedded,
        )

        # connect dialog signals to iface slots
        if service_type == "OGC:WMS/OGC:WMTS":
            ows_provider.addRasterLayer.connect(self.iface.addRasterLayer)
            conn_cmb = ows_provider.findChild(QWidget, "cmbConnections")
            connect = "btnConnect_clicked"
        elif service_type == "OGC:WFS":

            def addVectorLayer(path, name):
                self.iface.addVectorLayer(path, name, "WFS")

            ows_provider.addVectorLayer.connect(addVectorLayer)
            conn_cmb = ows_provider.findChild(QWidget, "cmbConnections")
            connect = "connectToServer"
        elif service_type == "OGC:WCS":
            ows_provider.addRasterLayer.connect(self.iface.addRasterLayer)
            conn_cmb = ows_provider.findChild(QWidget, "mConnectionsComboBox")
            connect = "mConnectButton_clicked"
        elif service_type == "ESRI:ArcGIS:FeatureServer":

            def addAfsLayer(path, name):
                self.iface.addVectorLayer(path, name, "afs")

            ows_provider.addVectorLayer.connect(addAfsLayer)
            conn_cmb = ows_provider.findChild(QComboBox)
            connect = "connectToServer"

        ows_provider.setModal(False)
        ows_provider.show()

        # open provider dialogue against added OWS
        index = conn_cmb.findText(sname)
        if index > -1:
            conn_cmb.setCurrentIndex(index)
            # only for wfs
            if service_type == "OGC:WFS":
                ows_provider.cmbConnections_activated(index)
            elif service_type == "ESRI:ArcGIS:FeatureServer":
                ows_provider.cmbConnections_activated(index)
        getattr(ows_provider, connect)()

    def add_gis_file(self):
        """add GIS file from result"""
        item = self.treeRecords.currentItem()

        if not item:
            return

        item_data = json.loads(get_item_data(item, "link"))
        gis_file = item_data["gis_file"]

        title = item_data["title"]

        layer = self.iface.addVectorLayer(gis_file, title, "ogr")
        if not layer:
            self.iface.messageBar().pushWarning(None, "Layer failed to load!")

    def show_metadata(self):
        """show record metadata"""

        if not self.treeRecords.selectedItems():
            return

        item = self.treeRecords.currentItem()
        if not item:
            return

        identifier = get_item_data(item, "identifier")

        auth = None

        if self.disable_ssl_verification:
            try:
                auth = Authentication(verify=False)
            except NameError:
                pass

        try:
            with OverrideCursor(Qt.CursorShape.WaitCursor):
                cat = get_catalog_service(
                    self.catalog_url,  # spellok
                    catalog_type=self.catalog_type,
                    timeout=self.timeout,
                    username=self.catalog_username or None,
                    password=self.catalog_password or None,
                    auth=auth,
                )
                record = cat.get_record(identifier)
                if cat.type == "OGC API - Records":
                    record["url"] = cat.conn.request
                elif cat.type == "OGC CSW 2.0.2":
                    record.url = cat.conn.request

        except Exception as err:
            QMessageBox.warning(
                self,
                self.tr("GetRecords error"),
                self.tr("Error getting response: {0}").format(err),
            )
            return
        except KeyError as err:
            QMessageBox.warning(
                self,
                self.tr("Record parsing error"),
                self.tr("Unable to locate record identifier: {0}").format(err),
            )
            return

        crd = RecordDialog()
        metadata = render_template(
            "en", self.context, record, self.catalog.record_info_template
        )

        style = QgsApplication.reportStyleSheet()
        crd.textMetadata.document().setDefaultStyleSheet(style)
        crd.textMetadata.setHtml(metadata)
        crd.exec()

    def show_api(self):
        """show API request / response"""

        crd = APIRequestResponseDialog(
            self.catalog.request, self.catalog.response, self.catalog.format
        )
        crd.exec()

    def reset_buttons(self, services=True, api=True, navigation=True):
        """Convenience function to disable WMS/WMTS|WFS|WCS buttons"""

        if services:
            self.tbAddData.setEnabled(False)
            self.mActionAddWms.setEnabled(False)
            self.mActionAddWfs.setEnabled(False)
            self.mActionAddWcs.setEnabled(False)
            self.mActionAddAms.setEnabled(False)
            self.mActionAddAfs.setEnabled(False)
            self.mActionAddGisFile.setEnabled(False)

        if api:
            self.btnViewRawAPIResponse.setEnabled(False)

        if navigation:
            self.btnFirst.setEnabled(False)
            self.btnPrev.setEnabled(False)
            self.btnNext.setEnabled(False)
            self.btnLast.setEnabled(False)

    def help(self):
        """launch help"""

        open_url(get_help_url())

    def reject(self):
        """back out of dialogue"""

        QDialog.reject(self)
        self.rubber_band.reset()

    def _get_catalog(self):
        """convenience function to init catalog wrapper"""

        auth = None

        if self.disable_ssl_verification:
            try:
                auth = Authentication(verify=False)
            except NameError:
                pass

        # connect to the server
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            try:
                self.catalog = get_catalog_service(
                    self.catalog_url,
                    catalog_type=self.catalog_type,
                    timeout=self.timeout,
                    username=self.catalog_username or None,
                    password=self.catalog_password or None,
                    auth=auth,
                )
                return True
            except Exception as err:
                msg = self.tr("Error connecting to service: {0}").format(err)

        QMessageBox.warning(self, self.tr("CSW Connection error"), msg)
        return False

    def install_proxy(self):
        """set proxy if one is set in QGIS network settings"""

        proxy_node = QgsSettingsTree.node("proxy")

        # initially support HTTP for now
        if proxy_node.childSetting("proxy-enabled").valueAsVariant():
            if proxy_node.childSetting("proxy-type").valueAsVariant() == "HttpProxy":
                ptype = "http"
            else:
                return

            user = proxy_node.childSetting("proxy-user").valueAsVariant() or ""
            password = proxy_node.childSetting("proxy-password").valueAsVariant() or ""
            host = proxy_node.childSetting("proxy-host").valueAsVariant() or ""
            port = proxy_node.childSetting("proxy-port").valueAsVariant() or ""

            proxy_up = ""
            proxy_port = ""

            if all([user != "", password != ""]):
                proxy_up = f"{user}:{password}@"

            if port != "":
                proxy_port = ":%s" % port

            conn = f"{ptype}://{proxy_up}{host}{proxy_port}"
            install_opener(build_opener(ProxyHandler({ptype: conn})))


def save_connections():
    """save servers to list"""

    ManageConnectionsDialog(0).exec()


def get_item_data(item, field):
    """return identifier for a QTreeWidgetItem"""

    return item.data(_get_field_value(field), 32)


def set_item_data(item, field, value):
    """set identifier for a QTreeWidgetItem"""

    item.setData(_get_field_value(field), 32, value)


def _get_field_value(field):
    """convenience function to return field value integer"""

    value = 0

    if field == "identifier":
        value = 0
    if field == "link":
        value = 1

    return value



"""
***************************************************************************
    TinInterpolation.py
    ---------------------
    Date                 : October 2016
    Copyright            : (C) 2016 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "October 2016"
__copyright__ = "(C) 2016, Alexander Bruy"

import math
import os

from qgis.analysis import QgsGridFileWriter, QgsInterpolator, QgsTinInterpolator
from qgis.core import (
    QgsCoordinateReferenceSystem,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterExtent,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterDestination,
    QgsProcessingUtils,
    QgsWkbTypes,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm
from processing.algs.qgis.ui.InterpolationWidgets import (
    ParameterInterpolationData,
    ParameterPixelSize,
)

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class TinInterpolation(QgisAlgorithm):
    INTERPOLATION_DATA = "INTERPOLATION_DATA"
    METHOD = "METHOD"
    PIXEL_SIZE = "PIXEL_SIZE"
    COLUMNS = "COLUMNS"
    ROWS = "ROWS"
    EXTENT = "EXTENT"
    OUTPUT = "OUTPUT"
    TRIANGULATION = "TRIANGULATION"

    def icon(self):
        return QIcon(os.path.join(pluginPath, "images", "interpolation.png"))

    def group(self):
        return self.tr("Interpolation")

    def groupId(self):
        return "interpolation"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.METHODS = [self.tr("Linear"), self.tr("Clough-Toucher (cubic)")]

        self.addParameter(
            ParameterInterpolationData(
                self.INTERPOLATION_DATA, self.tr("Input layer(s)")
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.METHOD,
                self.tr("Interpolation method"),
                options=self.METHODS,
                defaultValue=0,
            )
        )
        self.addParameter(
            QgsProcessingParameterExtent(self.EXTENT, self.tr("Extent"), optional=False)
        )
        pixel_size_param = ParameterPixelSize(
            self.PIXEL_SIZE,
            self.tr("Output raster size"),
            layersData=self.INTERPOLATION_DATA,
            extent=self.EXTENT,
            minValue=0.0,
            default=0.1,
        )
        self.addParameter(pixel_size_param)

        cols_param = QgsProcessingParameterNumber(
            self.COLUMNS,
            self.tr("Number of columns"),
            optional=True,
            minValue=0,
            maxValue=10000000,
        )
        cols_param.setFlags(
            cols_param.flags() | QgsProcessingParameterDefinition.Flag.FlagHidden
        )
        self.addParameter(cols_param)

        rows_param = QgsProcessingParameterNumber(
            self.ROWS,
            self.tr("Number of rows"),
            optional=True,
            minValue=0,
            maxValue=10000000,
        )
        rows_param.setFlags(
            rows_param.flags() | QgsProcessingParameterDefinition.Flag.FlagHidden
        )
        self.addParameter(rows_param)

        self.addParameter(
            QgsProcessingParameterRasterDestination(
                self.OUTPUT, self.tr("Interpolated")
            )
        )

        triangulation_file_param = QgsProcessingParameterFeatureSink(
            self.TRIANGULATION,
            self.tr("Triangulation"),
            type=QgsProcessing.SourceType.TypeVectorLine,
            optional=True,
        )
        triangulation_file_param.setCreateByDefault(False)
        self.addParameter(triangulation_file_param)

    def name(self):
        return "tininterpolation"

    def displayName(self):
        return self.tr("TIN interpolation")

    def shortDescription(self):
        return self.tr(
            "Generates a Triangulated Irregular Network (TIN) interpolation of a point vector layer."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm generates a Triangulated Irregular Network (TIN) interpolation of a point vector layer.\n"
            "With the TIN method you can create a surface formed by triangles of nearest neighbor points."
            "To do this, circumcircles around selected sample points are created and their intersections "
            "are connected to a network of non overlapping and as compact as possible triangles."
            "The resulting surfaces are not smooth.\n"
            "The algorithm creates both the raster layer of the interpolated values "
            "and the vector line layer with the triangulation boundaries."
        )

    def processAlgorithm(self, parameters, context, feedback):
        interpolationData = ParameterInterpolationData.parseValue(
            parameters[self.INTERPOLATION_DATA]
        )
        method = self.parameterAsEnum(parameters, self.METHOD, context)
        bbox = self.parameterAsExtent(parameters, self.EXTENT, context)
        pixel_size = self.parameterAsDouble(parameters, self.PIXEL_SIZE, context)
        output = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)

        columns = self.parameterAsInt(parameters, self.COLUMNS, context)
        rows = self.parameterAsInt(parameters, self.ROWS, context)
        if columns == 0:
            columns = max(math.ceil(bbox.width() / pixel_size), 1)
        if rows == 0:
            rows = max(math.ceil(bbox.height() / pixel_size), 1)

        if interpolationData is None:
            raise QgsProcessingException(
                self.tr("You need to specify at least one input layer.")
            )

        layerData = []
        layers = []
        crs = QgsCoordinateReferenceSystem()
        for i, row in enumerate(interpolationData.split("::|::")):
            v = row.split("::~::")
            data = QgsInterpolator.LayerData()

            # need to keep a reference until interpolation is complete
            layer = QgsProcessingUtils.variantToSource(v[0], context)
            data.source = layer
            data.transformContext = context.transformContext()
            layers.append(layer)
            if not crs.isValid():
                crs = layer.sourceCrs()

            data.valueSource = QgsInterpolator.ValueSource(int(v[1]))
            data.interpolationAttribute = int(v[2])
            if (
                data.valueSource == QgsInterpolator.ValueSource.Attribute
                and data.interpolationAttribute == -1
            ):
                raise QgsProcessingException(
                    self.tr(
                        "Layer {} is set to use a value attribute, but no attribute was set"
                    ).format(i + 1)
                )

            data.sourceType = QgsInterpolator.SourceType(int(v[3]))
            layerData.append(data)

        if method == 0:
            interpolationMethod = QgsTinInterpolator.TinInterpolation.Linear
        else:
            interpolationMethod = QgsTinInterpolator.TinInterpolation.CloughTocher

        (triangulation_sink, triangulation_dest_id) = self.parameterAsSink(
            parameters,
            self.TRIANGULATION,
            context,
            QgsTinInterpolator.triangulationFields(),
            QgsWkbTypes.Type.LineString,
            crs,
        )

        interpolator = QgsTinInterpolator(layerData, interpolationMethod, feedback)
        if triangulation_sink is not None:
            interpolator.setTriangulationSink(triangulation_sink)

        writer = QgsGridFileWriter(interpolator, output, bbox, columns, rows)

        writer.writeFile(feedback)
        if triangulation_sink:
            triangulation_sink.finalize()
        return {self.OUTPUT: output, self.TRIANGULATION: triangulation_dest_id}



"""
***************************************************************************
    AlgorithmDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import datetime
import time

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsProcessingAlgorithm,
    QgsProcessingAlgRunnerTask,
    QgsProcessingFeatureSourceDefinition,
    QgsProcessingOutputHtml,
    QgsProxyProgressTask,
)
from qgis.gui import (
    QgsGui,
    QgsProcessingAlgorithmDialogBase,
    QgsProcessingContextGenerator,
    QgsProcessingParametersGenerator,
)
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QColor, QPalette
from qgis.PyQt.QtWidgets import QDialogButtonBox, QMessageBox, QPushButton
from qgis.utils import iface

from processing.core.ProcessingConfig import ProcessingConfig
from processing.core.ProcessingResults import resultsList
from processing.gui.AlgorithmDialogBase import AlgorithmDialogBase
from processing.gui.AlgorithmExecutor import execute, execute_in_place, executeIterating
from processing.gui.BatchAlgorithmDialog import BatchAlgorithmDialog
from processing.gui.ParametersPanel import ParametersPanel
from processing.gui.Postprocessing import handleAlgorithmResults
from processing.tools import dataobjects


class AlgorithmDialog(QgsProcessingAlgorithmDialogBase):
    def __init__(self, alg, in_place=False, parent=None):
        super().__init__(parent)

        self.feedback_dialog = None
        self.in_place = in_place
        self.active_layer = iface.activeLayer() if self.in_place else None

        self.context = None
        self.feedback = None
        self.history_log_id = None
        self.history_details = {}

        self.setAlgorithm(alg)
        self.setMainWidget(self.getParametersPanel(alg, self))

        if not self.in_place:
            self.runAsBatchButton = QPushButton(
                QCoreApplication.translate("AlgorithmDialog", "Run as Batch Process…")
            )
            self.runAsBatchButton.clicked.connect(self.runAsBatch)
            self.buttonBox().addButton(
                self.runAsBatchButton, QDialogButtonBox.ButtonRole.ResetRole
            )  # reset role to ensure left alignment
        else:
            in_place_input_parameter_name = "INPUT"
            if hasattr(alg, "inputParameterName"):
                in_place_input_parameter_name = alg.inputParameterName()

            self.mainWidget().setParameters(
                {in_place_input_parameter_name: self.active_layer}
            )

            self.runAsBatchButton = None

            if self.active_layer:
                self.updateRunButtonText()
                self.active_layer.selectionChanged.connect(self.updateRunButtonText)
                self.active_layer.willBeDeleted.connect(self.close)
                self.setWindowTitle(
                    self.windowTitle() + " | " + self.active_layer.name()
                )

        self.updateRunButtonVisibility()

    def updateRunButtonText(self):
        has_selection = self.active_layer.selectedFeatureCount() > 0
        self.buttonBox().button(QDialogButtonBox.StandardButton.Ok).setText(
            QCoreApplication.translate("AlgorithmDialog", "Modify Selected Features")
            if has_selection
            else QCoreApplication.translate("AlgorithmDialog", "Modify All Features")
        )

    def getParametersPanel(self, alg, parent):
        panel = ParametersPanel(parent, alg, self.in_place, self.active_layer)
        return panel

    def runAsBatch(self):
        self.close()
        dlg = BatchAlgorithmDialog(self.algorithm().create(), parent=iface.mainWindow())
        dlg.show()
        dlg.exec()

    def resetAdditionalGui(self):
        if not self.in_place:
            self.runAsBatchButton.setEnabled(True)

    def blockAdditionalControlsWhileRunning(self):
        if not self.in_place:
            self.runAsBatchButton.setEnabled(False)

    def setParameters(self, parameters):
        self.mainWidget().setParameters(parameters)

    def flag_invalid_parameter_value(self, message: str, widget):
        """
        Highlights a parameter with an invalid value
        """
        try:
            self.buttonBox().accepted.connect(lambda w=widget: w.setPalette(QPalette()))
            palette = widget.palette()
            palette.setColor(QPalette.ColorRole.Base, QColor(255, 255, 0))
            widget.setPalette(palette)
        except:
            pass
        self.messageBar().clearWidgets()
        self.messageBar().pushMessage(
            "",
            self.tr("Wrong or missing parameter value: {0}").format(message),
            level=Qgis.MessageLevel.Warning,
            duration=5,
        )

    def flag_invalid_output_extension(self, message: str, widget):
        """
        Highlights a parameter with an invalid output extension
        """
        try:
            self.buttonBox().accepted.connect(lambda w=widget: w.setPalette(QPalette()))
            palette = widget.palette()
            palette.setColor(QPalette.ColorRole.Base, QColor(255, 255, 0))
            widget.setPalette(palette)
        except:
            pass
        self.messageBar().clearWidgets()
        self.messageBar().pushMessage(
            "", message, level=Qgis.MessageLevel.Warning, duration=5
        )

    def createProcessingParameters(
        self, flags=QgsProcessingParametersGenerator.Flags()
    ):
        if self.mainWidget() is None:
            return {}

        try:
            return self.mainWidget().createProcessingParameters(flags)
        except AlgorithmDialogBase.InvalidParameterValue as e:
            self.flag_invalid_parameter_value(e.parameter.description(), e.widget)
        except AlgorithmDialogBase.InvalidOutputExtension as e:
            self.flag_invalid_output_extension(e.message, e.widget)
        return {}

    def processingContext(self):
        if self.context is None:
            self.feedback = self.createFeedback()
            self.context = dataobjects.createContext(self.feedback)

        self.applyContextOverrides(self.context)
        return self.context

    def runAlgorithm(self):
        self.feedback = self.createFeedback()
        self.context = dataobjects.createContext(self.feedback)
        self.applyContextOverrides(self.context)
        self.algorithmAboutToRun.emit(self.context)

        checkCRS = ProcessingConfig.getSetting(ProcessingConfig.WARN_UNMATCHING_CRS)
        try:
            # messy as all heck, but we don't want to call the dialog's implementation of
            # createProcessingParameters as we want to catch the exceptions raised by the
            # parameter panel instead...
            parameters = (
                {}
                if self.mainWidget() is None
                else self.mainWidget().createProcessingParameters()
            )

            if checkCRS and not self.algorithm().validateInputCrs(
                parameters, self.context
            ):
                reply = QMessageBox.question(
                    self,
                    self.tr("Unmatching CRS's"),
                    self.tr(
                        "Parameters do not all use the same CRS. This can "
                        "cause unexpected results.\nDo you want to "
                        "continue?"
                    ),
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    QMessageBox.StandardButton.No,
                )
                if reply == QMessageBox.StandardButton.No:
                    return
            ok, msg = self.algorithm().checkParameterValues(parameters, self.context)
            if not ok:
                QMessageBox.warning(self, self.tr("Unable to execute algorithm"), msg)
                return

            self.blockControlsWhileRunning()
            self.setExecutedAnyResult(True)
            self.cancelButton().setEnabled(False)

            self.iterateParam = None

            for param in self.algorithm().parameterDefinitions():
                if (
                    isinstance(
                        parameters.get(param.name(), None),
                        QgsProcessingFeatureSourceDefinition,
                    )
                    and parameters[param.name()].flags
                    & QgsProcessingFeatureSourceDefinition.Flag.FlagCreateIndividualOutputPerInputFeature
                ):
                    self.iterateParam = param.name()
                    break

            self.clearProgress()
            self.feedback.pushVersionInfo(self.algorithm().provider())
            if (
                self.algorithm().provider()
                and self.algorithm().provider().warningMessage()
            ):
                self.feedback.reportError(self.algorithm().provider().warningMessage())

            self.feedback.pushInfo(
                QCoreApplication.translate(
                    "AlgorithmDialog", "Algorithm started at: {}"
                ).format(datetime.datetime.now().replace(microsecond=0).isoformat())
            )

            self.setInfo(
                QCoreApplication.translate(
                    "AlgorithmDialog", "<b>Algorithm '{0}' starting&hellip;</b>"
                ).format(self.algorithm().displayName()),
                escapeHtml=False,
            )

            self.feedback.pushInfo(self.tr("Input parameters:"))
            display_params = []
            for k, v in parameters.items():
                display_params.append(
                    "'"
                    + k
                    + "' : "
                    + self.algorithm()
                    .parameterDefinition(k)
                    .valueAsPythonString(v, self.context)
                )
            self.feedback.pushCommandInfo("{ " + ", ".join(display_params) + " }")
            self.feedback.pushInfo("")
            start_time = time.time()

            def elapsed_time(start_time) -> str:
                delta_t = time.time() - start_time
                hours = int(delta_t / 3600)
                minutes = int((delta_t % 3600) / 60)
                seconds = delta_t - hours * 3600 - minutes * 60

                str_hours = [self.tr("hour"), self.tr("hours")][hours > 1]
                str_minutes = [self.tr("minute"), self.tr("minutes")][minutes > 1]
                str_seconds = [self.tr("second"), self.tr("seconds")][seconds != 1]

                if hours > 0:
                    elapsed = f"{delta_t:0.2f} {str_seconds} ({hours} {str_hours} {minutes} {str_minutes} {seconds:0.0f} {str_seconds})"
                elif minutes > 0:
                    elapsed = f"{delta_t:0.2f} {str_seconds} ({minutes} {str_minutes} {seconds:0.0f} {str_seconds})"
                else:
                    elapsed = f"{delta_t:0.2f} {str_seconds}"

                return elapsed

            if self.iterateParam:
                # Make sure the Log tab is visible before executing the algorithm
                try:
                    self.showLog()
                    self.repaint()
                except:
                    pass

                self.cancelButton().setEnabled(
                    self.algorithm().flags() & QgsProcessingAlgorithm.Flag.FlagCanCancel
                )
                if executeIterating(
                    self.algorithm(),
                    parameters,
                    self.iterateParam,
                    self.context,
                    self.feedback,
                ):
                    self.feedback.pushInfo(
                        self.tr("Execution completed in {}").format(
                            elapsed_time(start_time)
                        )
                    )
                    self.cancelButton().setEnabled(False)
                    self.finish(True, parameters, self.context, self.feedback)
                else:
                    self.cancelButton().setEnabled(False)
                    self.resetGui()
            else:
                self.history_details = {
                    "python_command": self.algorithm().asPythonCommand(
                        parameters, self.context
                    ),
                    "algorithm_id": self.algorithm().id(),
                    "parameters": self.algorithm().asMap(parameters, self.context),
                }
                process_command, command_ok = self.algorithm().asQgisProcessCommand(
                    parameters, self.context
                )
                if command_ok:
                    self.history_details["process_command"] = process_command
                self.history_log_id, _ = QgsGui.historyProviderRegistry().addEntry(
                    "processing", self.history_details
                )

                QgsGui.instance().processingRecentAlgorithmLog().push(
                    self.algorithm().id()
                )
                self.cancelButton().setEnabled(
                    self.algorithm().flags() & QgsProcessingAlgorithm.Flag.FlagCanCancel
                )

                def on_complete(ok, results):
                    if ok:
                        self.feedback.pushInfo(
                            self.tr("Execution completed in {}").format(
                                elapsed_time(start_time)
                            )
                        )
                        self.feedback.pushFormattedResults(
                            self.algorithm(), self.context, results
                        )
                    else:
                        self.feedback.reportError(
                            self.tr("Execution failed after {}").format(
                                elapsed_time(start_time)
                            )
                        )
                    self.feedback.pushInfo("")

                    if self.history_log_id is not None:
                        # can't deepcopy this!
                        self.history_details["results"] = {
                            k: v for k, v in results.items() if k != "CHILD_INPUTS"
                        }
                        self.history_details["log"] = self.feedback.htmlLog()

                        QgsGui.historyProviderRegistry().updateEntry(
                            self.history_log_id, self.history_details
                        )

                    if self.feedback_dialog is not None:
                        self.feedback_dialog.close()
                        self.feedback_dialog.deleteLater()
                        self.feedback_dialog = None

                    self.cancelButton().setEnabled(False)

                    self.finish(
                        ok, results, self.context, self.feedback, in_place=self.in_place
                    )

                    self.feedback = None
                    self.context = None

                if not self.in_place and not (
                    self.algorithm().flags()
                    & QgsProcessingAlgorithm.Flag.FlagNoThreading
                ):
                    # Make sure the Log tab is visible before executing the algorithm
                    self.showLog()

                    task = QgsProcessingAlgRunnerTask(
                        self.algorithm(), parameters, self.context, self.feedback
                    )
                    if task.isCanceled():
                        on_complete(False, {})
                    else:
                        task.executed.connect(on_complete)
                        self.setCurrentTask(task)
                else:
                    self.proxy_progress = QgsProxyProgressTask(
                        QCoreApplication.translate(
                            "AlgorithmDialog", "Executing “{}”"
                        ).format(self.algorithm().displayName())
                    )
                    QgsApplication.taskManager().addTask(self.proxy_progress)
                    self.feedback.progressChanged.connect(
                        self.proxy_progress.setProxyProgress
                    )
                    self.feedback_dialog = self.createProgressDialog()
                    self.feedback_dialog.show()
                    if self.in_place:
                        ok, results = execute_in_place(
                            self.algorithm(), parameters, self.context, self.feedback
                        )
                    else:
                        ok, results = execute(
                            self.algorithm(), parameters, self.context, self.feedback
                        )
                    self.feedback.progressChanged.disconnect()
                    self.proxy_progress.finalize(ok)
                    on_complete(ok, results)

        except AlgorithmDialogBase.InvalidParameterValue as e:
            self.flag_invalid_parameter_value(e.parameter.description(), e.widget)
        except AlgorithmDialogBase.InvalidOutputExtension as e:
            self.flag_invalid_output_extension(e.message, e.widget)

    def finish(self, successful, result, context, feedback, in_place=False):
        keepOpen = not successful or ProcessingConfig.getSetting(
            ProcessingConfig.KEEP_DIALOG_OPEN
        )
        generated_html_outputs = False

        if not in_place and self.iterateParam is None:
            # add html results to results dock
            for out in self.algorithm().outputDefinitions():
                if (
                    isinstance(out, QgsProcessingOutputHtml)
                    and out.name() in result
                    and result[out.name()]
                ):
                    resultsList.addResult(
                        icon=self.algorithm().icon(),
                        name=out.description(),
                        timestamp=time.localtime(),
                        result=result[out.name()],
                    )
                    generated_html_outputs = True
            if not handleAlgorithmResults(self.algorithm(), context, feedback, result):
                self.resetGui()
                return

        self.setExecuted(True)
        self.setResults(result)
        self.setInfo(
            self.tr("Algorithm '{0}' finished").format(self.algorithm().displayName()),
            escapeHtml=False,
        )
        self.algorithmFinished.emit(successful, result)

        if not in_place and not keepOpen:
            self.close()
        else:
            self.resetGui()
            if generated_html_outputs:
                self.setInfo(
                    self.tr(
                        "HTML output has been generated by this algorithm."
                        "\nOpen the results dialog to check it."
                    ),
                    escapeHtml=False,
                )



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtWidgets import QApplication, QTextBrowser
from qgis.utils import OverrideCursor

from .db_plugins.plugin import BaseError, DbError, DBPlugin, Schema, Table
from .dlg_db_error import DlgDbError
from .gui_utils import GuiUtils


class InfoViewer(QTextBrowser):
    def __init__(self, parent=None):
        QTextBrowser.__init__(self, parent)
        self.setOpenLinks(False)

        self.item = None
        self.dirty = False

        self._clear()
        self._showPluginInfo()

        self.anchorClicked.connect(self._linkClicked)

    def _linkClicked(self, url):
        if self.item is None:
            return

        if url.scheme() == "action":
            with OverrideCursor(Qt.CursorShape.WaitCursor):
                try:
                    if self.item.runAction(url.path()):
                        self.refresh()
                except BaseError as e:
                    DlgDbError.showError(e, self)

    def refresh(self):
        self.setDirty(True)
        self.showInfo(self.item)

    def showInfo(self, item):
        if item == self.item and not self.dirty:
            return
        self._clear()
        if item is None:
            return

        if isinstance(item, DBPlugin):
            self._showDatabaseInfo(item)
        elif isinstance(item, Schema):
            self._showSchemaInfo(item)
        elif isinstance(item, Table):
            self._showTableInfo(item)
        else:
            return

        self.item = item
        item.aboutToChange.connect(self.setDirty)

    def setDirty(self, val=True):
        self.dirty = val

    def _clear(self):
        if self.item is not None:
            # skip exception on RuntimeError fixes #6892
            try:
                self.item.aboutToChange.disconnect(self.setDirty)
            except RuntimeError:
                pass
        self.item = None
        self.dirty = False

        self.item = None
        self.setHtml("")

    def _showPluginInfo(self):
        from .db_plugins import getDbPluginErrors

        html = (
            '<div style="background-color:rgba(255,255,95,0.3);"><h1>&nbsp;'
            + self.tr("DB Manager")
            + "</h1></div>"
        )
        html += '<div style="margin-left:8px;">'
        for msg in getDbPluginErrors():
            html += "<p>%s" % msg
        self.setHtml(html)

    def _showDatabaseInfo(self, connection):
        html = (
            '<div style="background-color:rgba(120,255,100,0.3);"><h1>&nbsp;%s</h1></div>'
            % connection.connectionName()
        )
        html += '<div style="margin-left:8px;">'
        try:
            if connection.database() is None:
                html += connection.info().toHtml()
            else:
                html += connection.database().info().toHtml()
        except DbError as e:
            html += '<p style="color:red">%s</p>' % str(e).replace("\n", "<br>")
        html += "</div>"
        self.setHtml(html)

    def _showSchemaInfo(self, schema):
        html = (
            '<div style="background-color:rgba(255,100,100,0.3);"><h1>&nbsp;%s</h1></div>'
            % schema.name
        )
        html += '<div style="margin-left:8px;">'
        try:
            html += schema.info().toHtml()
        except DbError as e:
            html += '<p style="color:red">%s</p>' % str(e).replace("\n", "<br>")
        html += "</div>"
        self.setHtml(html)

    def _showTableInfo(self, table):
        html = (
            '<div style="background-color:rgba(100,100,255,0.3)"><h1>&nbsp;%s</h1></div>'
            % table.name
        )
        html += '<div style="margin-left:8px;">'
        try:
            html += table.info().toHtml()
        except DbError as e:
            html += '<p style="color:red">%s</p>' % str(e).replace("\n", "<br>")
        html += "</div>"
        self.setHtml(html)
        return True

    def setHtml(self, html):
        # convert special tags
        warning_icon_path = GuiUtils.get_pixmap_path("warning-20px")
        html = str(html).replace(
            "<warning>", f'<img src="{warning_icon_path}">&nbsp;&nbsp; '
        )

        # add default style
        html = (
            """
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
<style type="text/css">
        .section { margin-top: 25px; }
        table th { background-color: palette(midlight); color: palette(shadow); }
        table td { background-color: palette(light); }
        table th, table td { padding: 0px 10px; }
        table td { padding-right: 20px; }
        .underline { text-decoration:underline; }
</style>
</head>
<body>
%s <br>
</body>
</html>
"""
            % html
        )

        # print(">>>>>\n", html, "\n<<<<<<")
        return QTextBrowser.setHtml(self, html)



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

The content of this file is based on
- PG_Manager by Martin Dobias (GPLv2 license)
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os
from hashlib import md5

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsMapLayerType,
    QgsProject,
    QgsSettings,
    QgsTask,
)
from qgis.gui import QgsGui
from qgis.PyQt import uic
from qgis.PyQt.Qsci import QsciAPIs, QsciScintilla
from qgis.PyQt.QtCore import QCoreApplication, QDir, Qt, pyqtSignal
from qgis.PyQt.QtGui import (
    QClipboard,
    QCursor,
    QIcon,
    QKeySequence,
    QStandardItem,
    QStandardItemModel,
)
from qgis.PyQt.QtWidgets import (
    QAction,
    QApplication,
    QDialog,
    QFileDialog,
    QInputDialog,
    QMessageBox,
    QStyledItemDelegate,
    QTableWidgetItem,
    QWidget,
)
from qgis.utils import OverrideCursor

from .db_plugins.plugin import BaseError
from .db_plugins.postgis.plugin import PGDatabase
from .dlg_db_error import DlgDbError
from .dlg_query_builder import QueryBuilderDlg
from .gui_utils import GuiUtils

try:
    from qgis.gui import QgsCodeEditorSQL  # NOQA
except:
    from qgis import gui

    from .sqledit import SqlEdit

    gui.QgsCodeEditorSQL = SqlEdit

Ui_Dialog, _ = uic.loadUiType(GuiUtils.get_ui_file_path("DlgSqlWindow.ui"))


import re


def check_comments_in_sql(raw_sql_input):
    lines = []
    for line in raw_sql_input.splitlines():
        if not line.strip().startswith("--"):
            if "--" in line:
                comments = re.finditer(r"--", line)
                comment_positions = [match.start() for match in comments]
                identifiers = re.finditer(r'"(?:[^"]|"")*"', line)
                quotes = re.finditer(r"'(?:[^']|'')*'", line)
                quote_positions = []
                for match in identifiers:
                    quote_positions.append((match.start(), match.end()))
                for match in quotes:
                    quote_positions.append((match.start(), match.end()))
                unquoted_comments = comment_positions.copy()
                for comment in comment_positions:
                    for quote_position in quote_positions:
                        if comment >= quote_position[0] and comment < quote_position[1]:
                            unquoted_comments.remove(comment)
                if len(unquoted_comments) > 0:
                    lines.append(line[: unquoted_comments[0]])
                else:
                    lines.append(line)
            else:
                lines.append(line)
    sql = " ".join(lines)
    return sql.strip()


class DlgSqlWindow(QWidget, Ui_Dialog):
    nameChanged = pyqtSignal(str)
    QUERY_HISTORY_LIMIT = 20
    hasChanged = False

    def __init__(self, iface, db, parent=None):
        QWidget.__init__(self, parent)
        self.mainWindow = parent
        self.iface = iface
        self.db = db
        self.dbType = db.connection().typeNameString()
        self.connectionName = db.connection().connectionName()
        self.filter = ""
        self.modelAsync = None
        self.allowMultiColumnPk = isinstance(
            db, PGDatabase
        )  # at the moment only PostgreSQL allows a primary key to span multiple columns, SpatiaLite doesn't
        self.aliasSubQuery = isinstance(
            db, PGDatabase
        )  # only PostgreSQL requires subqueries to be aliases
        self.setupUi(self)
        self.setWindowTitle(
            self.tr("{0} - {1} [{2}]").format(
                self.windowTitle(), self.connectionName, self.dbType
            )
        )

        self.defaultLayerName = self.tr("QueryLayer")

        if self.allowMultiColumnPk:
            self.uniqueColumnCheck.setText(self.tr("Column(s) with unique values"))
        else:
            self.uniqueColumnCheck.setText(self.tr("Column with unique values"))

        self.editSql.setFocus()
        self.editSql.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.editSql.setLineNumbersVisible(True)
        self.initCompleter()
        self.editSql.textChanged.connect(lambda: self.setHasChanged(True))

        settings = QgsSettings()
        self.history = settings.value(
            "DB_Manager/queryHistory/" + self.dbType, {self.connectionName: []}
        )
        if self.connectionName not in self.history:
            self.history[self.connectionName] = []

        self.queryHistoryWidget.setVisible(False)
        self.queryHistoryTableWidget.verticalHeader().hide()
        self.queryHistoryTableWidget.doubleClicked.connect(self.insertQueryInEditor)
        self.populateQueryHistory()
        self.btnQueryHistory.toggled.connect(self.showHideQueryHistory)

        self.btnCancel.setEnabled(False)
        self.btnCancel.clicked.connect(self.executeSqlCanceled)
        self.btnCancel.setShortcut(QKeySequence.StandardKey.Cancel)
        self.progressBar.setEnabled(False)
        self.progressBar.setRange(0, 100)
        self.progressBar.setValue(0)
        self.progressBar.setFormat("")
        self.progressBar.setAlignment(Qt.AlignmentFlag.AlignCenter)

        # allow copying results
        copyAction = QAction("copy", self)
        self.viewResult.addAction(copyAction)
        copyAction.setShortcuts(QKeySequence.StandardKey.Copy)

        copyAction.triggered.connect(self.copySelectedResults)

        self.btnExecute.clicked.connect(self.executeSql)
        self.btnSetFilter.clicked.connect(self.setFilter)
        self.btnClear.clicked.connect(self.clearSql)

        self.presetStore.clicked.connect(self.storePreset)
        self.presetSaveAsFile.clicked.connect(self.saveAsFilePreset)
        self.presetLoadFile.clicked.connect(self.loadFilePreset)
        self.presetDelete.clicked.connect(self.deletePreset)
        self.presetCombo.textActivated.connect(self.loadPreset)
        self.presetCombo.textActivated.connect(self.presetName.setText)

        self.updatePresetsCombobox()

        self.geomCombo.setEditable(True)
        self.geomCombo.lineEdit().setReadOnly(True)

        self.uniqueCombo.setEditable(True)
        self.uniqueCombo.lineEdit().setReadOnly(True)
        self.uniqueModel = QStandardItemModel(self.uniqueCombo)
        self.uniqueCombo.setModel(self.uniqueModel)
        if self.allowMultiColumnPk:
            self.uniqueCombo.setItemDelegate(QStyledItemDelegate())
            self.uniqueModel.itemChanged.connect(
                self.uniqueChanged
            )  # react to the (un)checking of an item
            self.uniqueCombo.lineEdit().textChanged.connect(
                self.uniqueTextChanged
            )  # there are other events that change the displayed text and some of them can not be caught directly

        # hide the load query as layer if feature is not supported
        self._loadAsLayerAvailable = self.db.connector.hasCustomQuerySupport()
        self.loadAsLayerGroup.setVisible(self._loadAsLayerAvailable)
        if self._loadAsLayerAvailable:
            self.layerTypeWidget.hide()  # show if load as raster is supported
            self.loadLayerBtn.clicked.connect(self.loadSqlLayer)
            self.getColumnsBtn.clicked.connect(self.fillColumnCombos)
            self.loadAsLayerGroup.toggled.connect(self.loadAsLayerToggled)
            self.loadAsLayerToggled(False)

        self._createViewAvailable = self.db.connector.hasCreateSpatialViewSupport()
        self.btnCreateView.setVisible(self._createViewAvailable)
        if self._createViewAvailable:
            self.btnCreateView.clicked.connect(self.createView)

        self.queryBuilderFirst = True
        self.queryBuilderBtn.setIcon(GuiUtils.get_icon("sql"))
        self.queryBuilderBtn.clicked.connect(self.displayQueryBuilder)

        self.presetName.textChanged.connect(self.nameChanged)

    def insertQueryInEditor(self, item):
        sql = item.data(Qt.ItemDataRole.DisplayRole)
        self.editSql.insertText(sql)

    def showHideQueryHistory(self, visible):
        self.queryHistoryWidget.setVisible(visible)

    def populateQueryHistory(self):
        self.queryHistoryTableWidget.clearContents()
        self.queryHistoryTableWidget.setRowCount(0)
        dictlist = self.history[self.connectionName]

        if not dictlist:
            return

        for i in range(len(dictlist)):
            self.queryHistoryTableWidget.insertRow(0)
            queryItem = QTableWidgetItem(dictlist[i]["query"])
            rowsItem = QTableWidgetItem(str(dictlist[i]["rows"]))
            durationItem = QTableWidgetItem(str(dictlist[i]["secs"]))
            self.queryHistoryTableWidget.setItem(0, 0, queryItem)
            self.queryHistoryTableWidget.setItem(0, 1, rowsItem)
            self.queryHistoryTableWidget.setItem(0, 2, durationItem)

        self.queryHistoryTableWidget.resizeColumnsToContents()
        self.queryHistoryTableWidget.resizeRowsToContents()

    def writeQueryHistory(self, sql, affectedRows, secs):
        if len(self.history[self.connectionName]) >= self.QUERY_HISTORY_LIMIT:
            self.history[self.connectionName].pop(0)

        settings = QgsSettings()
        self.history[self.connectionName].append(
            {"query": sql, "rows": affectedRows, "secs": secs}
        )
        settings.setValue("DB_Manager/queryHistory/" + self.dbType, self.history)

        self.populateQueryHistory()

    def getQueryHash(self, name):
        return "q%s" % md5(name.encode("utf8")).hexdigest()

    def updatePresetsCombobox(self):
        self.presetCombo.clear()

        queries = QgsGui.storedQueryManager().allQueries()
        for query in queries:
            self.presetCombo.addItem(query.name, query.definition)

        self.presetCombo.setCurrentIndex(-1)

    def storePreset(self):
        query = self._getSqlQuery()
        if query == "":
            return
        name = str(self.presetName.text())

        QgsGui.storedQueryManager().storeQuery(
            name, query, Qgis.QueryStorageBackend.CurrentProject
        )
        self.updatePresetsCombobox()

        index = self.presetCombo.findText(name)
        if index != -1:
            self.presetCombo.setCurrentIndex(index)

    def saveAsFilePreset(self):
        settings = QgsSettings()
        lastDir = settings.value("DB_Manager/lastDirSQLFIle", "")

        query = self.editSql.text()
        if query == "":
            return

        filename, _ = QFileDialog.getSaveFileName(
            self, self.tr("Save SQL Query"), lastDir, self.tr("SQL File (*.sql *.SQL)")
        )

        if filename:
            if not filename.lower().endswith(".sql"):
                filename += ".sql"

            with open(filename, "w") as f:
                f.write(query)
                lastDir = os.path.dirname(filename)
                settings.setValue("DB_Manager/lastDirSQLFile", lastDir)

    def loadFilePreset(self):
        settings = QgsSettings()
        lastDir = settings.value("DB_Manager/lastDirSQLFIle", "")

        filename, _ = QFileDialog.getOpenFileName(
            self,
            self.tr("Load SQL Query"),
            lastDir,
            self.tr("SQL File (*.sql *.SQL);;All Files (*)"),
        )

        if filename:
            with open(filename) as f:
                self.editSql.clear()
                for line in f:
                    self.editSql.insertText(line)
                lastDir = os.path.dirname(filename)
                settings.setValue("DB_Manager/lastDirSQLFile", lastDir)

    def deletePreset(self):
        name = self.presetCombo.currentText()
        QgsGui.storedQueryManager().removeQuery(
            name, Qgis.QueryStorageBackend.CurrentProject
        )
        self.updatePresetsCombobox()

    def loadPreset(self, _):
        query = self.presetCombo.currentData()
        self.editSql.setText(query)

    def loadAsLayerToggled(self, checked):
        self.loadAsLayerGroup.setChecked(checked)
        self.loadAsLayerWidget.setVisible(checked)
        if checked:
            self.fillColumnCombos()

    def clearSql(self):
        self.editSql.clear()
        self.editSql.setFocus()
        self.filter = ""
        self.setHasChanged(True)

    def updateUiWhileSqlExecution(self, status):
        if status:
            for i in range(0, self.mainWindow.tabs.count()):
                if i != self.mainWindow.tabs.currentIndex():
                    self.mainWindow.tabs.setTabEnabled(i, False)

            self.mainWindow.menuBar.setEnabled(False)
            self.mainWindow.toolBar.setEnabled(False)
            self.mainWindow.tree.setEnabled(False)

            for w in self.findChildren(QWidget):
                w.setEnabled(False)

            self.btnCancel.setEnabled(True)
            self.progressBar.setEnabled(True)
            self.progressBar.setRange(0, 0)
        else:
            for i in range(0, self.mainWindow.tabs.count()):
                if i != self.mainWindow.tabs.currentIndex():
                    self.mainWindow.tabs.setTabEnabled(i, True)

            self.mainWindow.refreshTabs()
            self.mainWindow.menuBar.setEnabled(True)
            self.mainWindow.toolBar.setEnabled(True)
            self.mainWindow.tree.setEnabled(True)

            for w in self.findChildren(QWidget):
                w.setEnabled(True)

            self.btnCancel.setEnabled(False)
            self.progressBar.setRange(0, 100)
            self.progressBar.setEnabled(False)

    def executeSqlCanceled(self):
        self.btnCancel.setEnabled(False)
        self.btnCancel.setText(QCoreApplication.translate("DlgSqlWindow", "Canceling…"))
        self.modelAsync.cancel()

    def executeSqlCompleted(self):
        self.updateUiWhileSqlExecution(False)

        with OverrideCursor(Qt.CursorShape.WaitCursor):
            if self.modelAsync.task.status() == QgsTask.TaskStatus.Complete:
                model = self.modelAsync.model
                self.showError(None)
                self.viewResult.setModel(model)
                self.lblResult.setText(
                    self.tr("{0} rows, {1:.3f} seconds").format(
                        model.affectedRows(), model.secs()
                    )
                )
                cols = self.viewResult.model().columnNames()
                quotedCols = [self.db.connector.quoteId(col) for col in cols]

                self.setColumnCombos(cols, quotedCols)

                self.writeQueryHistory(
                    self.modelAsync.task.sql, model.affectedRows(), model.secs()
                )
                self.update()
            elif not self.modelAsync.canceled:
                self.showError(self.modelAsync.error)

                self.uniqueModel.clear()
                self.geomCombo.clear()

            self.btnCancel.setText(self.tr("Cancel"))

    def executeSql(self):
        sql = self._getExecutableSqlQuery()
        if sql == "":
            return

        # delete the old model
        old_model = self.viewResult.model()
        self.viewResult.setModel(None)
        if old_model:
            old_model.deleteLater()

        try:
            self.modelAsync = self.db.sqlResultModelAsync(sql, self)
            self.modelAsync.done.connect(self.executeSqlCompleted)
            self.updateUiWhileSqlExecution(True)
            QgsApplication.taskManager().addTask(self.modelAsync.task)
        except Exception as e:
            self.showError(e)
            self.uniqueModel.clear()
            self.geomCombo.clear()
            return

    def showError(self, error):
        """Shows the error or hides it if error is None"""
        if error:
            self.viewResult.setVisible(False)
            self.errorText.setVisible(True)
            self.errorText.setText(error.msg)
            self.errorText.setWrapMode(QsciScintilla.WrapMode.WrapWord)
        else:
            self.viewResult.setVisible(True)
            self.errorText.setVisible(False)

    def _getSqlLayer(self, _filter):
        hasUniqueField = self.uniqueColumnCheck.checkState() == Qt.CheckState.Checked
        if hasUniqueField and self.allowMultiColumnPk:
            uniqueFieldName = ",".join(
                item.data()
                for item in self.uniqueModel.findItems("*", Qt.MatchFlag.MatchWildcard)
                if item.checkState() == Qt.CheckState.Checked
            )
        elif (
            hasUniqueField
            and not self.allowMultiColumnPk
            and self.uniqueCombo.currentIndex() >= 0
        ):
            uniqueFieldName = self.uniqueModel.item(
                self.uniqueCombo.currentIndex()
            ).data()
        else:
            uniqueFieldName = None
        hasGeomCol = self.hasGeometryCol.checkState() == Qt.CheckState.Checked
        if hasGeomCol:
            geomFieldName = self.geomCombo.currentText()
        else:
            geomFieldName = None

        query = self._getExecutableSqlQuery()
        if query == "":
            return None

        # remove a trailing ';' from query if present
        if query.strip().endswith(";"):
            query = query.strip()[:-1]

        layerType = (
            QgsMapLayerType.VectorLayer
            if self.vectorRadio.isChecked()
            else QgsMapLayerType.RasterLayer
        )

        # get a new layer name
        names = []
        for layer in list(QgsProject.instance().mapLayers().values()):
            names.append(layer.name())

        layerName = self.layerNameEdit.text()
        if layerName == "":
            layerName = self.defaultLayerName
        newLayerName = layerName
        index = 1
        while newLayerName in names:
            index += 1
            newLayerName = "%s_%d" % (layerName, index)

        # create the layer
        layer = self.db.toSqlLayer(
            query,
            geomFieldName,
            uniqueFieldName,
            newLayerName,
            layerType,
            self.avoidSelectById.isChecked(),
            _filter,
        )
        if layer.isValid():
            return layer
        else:
            e = BaseError(
                self.tr(
                    "There was an error creating the SQL layer, please check the logs for further information."
                )
            )
            DlgDbError.showError(e, self)
            return None

    def loadSqlLayer(self):
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            layer = self._getSqlLayer(self.filter)
            if layer is None:
                return

            QgsProject.instance().addMapLayers([layer], True)

    def fillColumnCombos(self):
        query = self._getExecutableSqlQuery()
        if query == "":
            return

        with OverrideCursor(Qt.CursorShape.WaitCursor):
            # remove a trailing ';' from query if present
            if query.strip().endswith(";"):
                query = query.strip()[:-1]

            # get all the columns
            quotedCols = []
            connector = self.db.connector
            if self.aliasSubQuery:
                # get a new alias
                aliasIndex = 0
                while True:
                    alias = "_subQuery__%d" % aliasIndex
                    escaped = re.compile('\\b("?)' + re.escape(alias) + "\\1\\b")
                    if not escaped.search(query):
                        break
                    aliasIndex += 1

                sql = f"SELECT * FROM ({str(query)}\n) AS {connector.quoteId(alias)} LIMIT 0"
            else:
                sql = "SELECT * FROM (%s\n) WHERE 1=0" % str(query)

            c = None
            try:
                c = connector._execute(None, sql)
                cols = connector._get_cursor_columns(c)
                for col in cols:
                    quotedCols.append(connector.quoteId(col))

            except BaseError as e:
                DlgDbError.showError(e, self)
                self.uniqueModel.clear()
                self.geomCombo.clear()
                return

            finally:
                if c:
                    c.close()
                    del c

            self.setColumnCombos(cols, quotedCols)

    def setColumnCombos(self, cols, quotedCols):
        # get sensible default columns. do this before sorting in case there's hints in the column order (e.g., id is more likely to be first)
        try:
            defaultGeomCol = next(
                col for col in cols if col in ["geom", "geometry", "the_geom", "way"]
            )
        except:
            defaultGeomCol = None
        try:
            defaultUniqueCol = [col for col in cols if "id" in col][0]
        except:
            defaultUniqueCol = None

        colNames = sorted(zip(cols, quotedCols))
        newItems = []
        uniqueIsFilled = False
        for col, quotedCol in colNames:
            item = QStandardItem(col)
            item.setData(quotedCol)
            item.setEnabled(True)
            item.setCheckable(self.allowMultiColumnPk)
            item.setSelectable(not self.allowMultiColumnPk)
            if self.allowMultiColumnPk:
                matchingItems = self.uniqueModel.findItems(col)
                if matchingItems:
                    item.setCheckState(matchingItems[0].checkState())
                    uniqueIsFilled = (
                        uniqueIsFilled
                        or matchingItems[0].checkState() == Qt.CheckState.Checked
                    )
                else:
                    item.setCheckState(Qt.CheckState.Unchecked)
            newItems.append(item)
        if self.allowMultiColumnPk:
            self.uniqueModel.clear()
            self.uniqueModel.appendColumn(newItems)
            self.uniqueChanged()
        else:
            previousUniqueColumn = self.uniqueCombo.currentText()
            self.uniqueModel.clear()
            self.uniqueModel.appendColumn(newItems)
            if self.uniqueModel.findItems(previousUniqueColumn):
                self.uniqueCombo.setEditText(previousUniqueColumn)
                uniqueIsFilled = True

        oldGeometryColumn = self.geomCombo.currentText()
        self.geomCombo.clear()
        self.geomCombo.addItems(cols)
        self.geomCombo.setCurrentIndex(
            self.geomCombo.findText(oldGeometryColumn, Qt.MatchFlag.MatchExactly)
        )

        # set sensible default columns if the columns are not already set
        try:
            if self.geomCombo.currentIndex() == -1:
                self.geomCombo.setCurrentIndex(cols.index(defaultGeomCol))
        except:
            pass
        items = self.uniqueModel.findItems(defaultUniqueCol)
        if items and not uniqueIsFilled:
            if self.allowMultiColumnPk:
                items[0].setCheckState(Qt.CheckState.Checked)
            else:
                self.uniqueCombo.setEditText(defaultUniqueCol)

    def copySelectedResults(self):
        if len(self.viewResult.selectedIndexes()) <= 0:
            return
        model = self.viewResult.model()

        # convert to string using tab as separator
        text = model.headerToString("\t")
        for idx in self.viewResult.selectionModel().selectedRows():
            text += "\n" + model.rowToString(idx.row(), "\t")

        QApplication.clipboard().setText(text, QClipboard.Mode.Selection)
        QApplication.clipboard().setText(text, QClipboard.Mode.Clipboard)

    def initCompleter(self):
        dictionary = None
        if self.db:
            dictionary = self.db.connector.getSqlDictionary()
        if not dictionary:
            # use the generic sql dictionary
            from .sql_dictionary import getSqlDictionary

            dictionary = getSqlDictionary()

        wordlist = []
        for value in dictionary.values():
            wordlist += value  # concat lists
        wordlist = list(set(wordlist))  # remove duplicates

        api = QsciAPIs(self.editSql.lexer())
        for word in wordlist:
            api.add(word)

        api.prepare()
        self.editSql.lexer().setAPIs(api)

    def displayQueryBuilder(self):
        dlg = QueryBuilderDlg(self.iface, self.db, self, reset=self.queryBuilderFirst)
        self.queryBuilderFirst = False
        r = dlg.exec()
        if r == QDialog.DialogCode.Accepted:
            self.editSql.setText(dlg.query)

    def createView(self):
        name, ok = QInputDialog.getText(
            None, self.tr("View Name"), self.tr("View name")
        )
        if ok:
            try:
                self.db.connector.createSpatialView(name, self._getExecutableSqlQuery())
            except BaseError as e:
                DlgDbError.showError(e, self)

    def _getSqlQuery(self):
        sql = self.editSql.selectedText()
        if len(sql) == 0:
            sql = self.editSql.text()
        return sql

    def _getExecutableSqlQuery(self):
        sql = self._getSqlQuery().strip()

        uncommented_sql = check_comments_in_sql(sql)
        uncommented_sql = uncommented_sql.rstrip(";")
        return uncommented_sql

    def uniqueChanged(self):
        # when an item is (un)checked, simply trigger an update of the combobox text
        self.uniqueTextChanged(None)

    def uniqueTextChanged(self, text):
        # Whenever there is new text displayed in the combobox, check if it is the correct one and if not, display the correct one.
        label = ", ".join(
            item.text()
            for item in self.uniqueModel.findItems("*", Qt.MatchFlag.MatchWildcard)
            if item.checkState() == Qt.CheckState.Checked
        )
        if text != label:
            self.uniqueCombo.setEditText(label)

    def setFilter(self):
        from qgis.gui import QgsQueryBuilder

        layer = self._getSqlLayer("")
        if not layer:
            return

        dlg = QgsQueryBuilder(layer)
        dlg.setSql(self.filter)
        if dlg.exec():
            self.filter = dlg.sql()
        layer.deleteLater()

    def setHasChanged(self, hasChanged):
        self.hasChanged = hasChanged

    def close(self):
        if self.hasChanged:
            ret = QMessageBox.question(
                self,
                self.tr("Unsaved Changes?"),
                self.tr("There are unsaved changes. Do you want to keep them?"),
                QMessageBox.StandardButton.Save
                | QMessageBox.StandardButton.Cancel
                | QMessageBox.StandardButton.Discard,
                QMessageBox.StandardButton.Cancel,
            )

            if ret == QMessageBox.StandardButton.Save:
                self.saveAsFilePreset()
                return True
            elif ret == QMessageBox.StandardButton.Discard:
                return True
            else:
                return False
        else:
            return True



"""
***************************************************************************
    pansharp.py
    ---------------------
    Date                 : March 2019
    Copyright            : (C) 2019 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "March 2019"
__copyright__ = "(C) 2019, Alexander Bruy"

import os

from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsRasterFileWriter,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils
from processing.tools.system import isWindows

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class pansharp(GdalAlgorithm):
    SPECTRAL = "SPECTRAL"
    PANCHROMATIC = "PANCHROMATIC"
    RESAMPLING = "RESAMPLING"
    OPTIONS = "OPTIONS"
    CREATION_OPTIONS = "CREATION_OPTIONS"
    EXTRA = "EXTRA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.methods = (
            (self.tr("Nearest Neighbour"), "nearest"),
            (self.tr("Bilinear (2x2 Kernel)"), "bilinear"),
            (self.tr("Cubic (4x4 Kernel)"), "cubic"),
            (self.tr("Cubic B-Spline (4x4 Kernel)"), "cubicspline"),
            (self.tr("Lanczos (6x6 Kernel)"), "lanczos"),
            (self.tr("Average"), "average"),
        )

        self.addParameter(
            QgsProcessingParameterRasterLayer(
                self.SPECTRAL, self.tr("Spectral dataset")
            )
        )
        self.addParameter(
            QgsProcessingParameterRasterLayer(
                self.PANCHROMATIC, self.tr("Panchromatic dataset")
            )
        )

        resampling_param = QgsProcessingParameterEnum(
            self.RESAMPLING,
            self.tr("Resampling algorithm"),
            options=[i[0] for i in self.methods],
            defaultValue=2,
        )
        resampling_param.setFlags(
            resampling_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(resampling_param)

        # backwards compatibility parameter
        # TODO QGIS 5: remove parameter and related logic
        options_param = QgsProcessingParameterString(
            self.OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        options_param.setFlags(
            options_param.flags() | QgsProcessingParameterDefinition.Flag.Hidden
        )
        options_param.setMetadata({"widget_wrapper": {"widget_type": "rasteroptions"}})
        self.addParameter(options_param)

        creation_options_param = QgsProcessingParameterString(
            self.CREATION_OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        creation_options_param.setFlags(
            creation_options_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        creation_options_param.setMetadata(
            {"widget_wrapper": {"widget_type": "rasteroptions"}}
        )
        self.addParameter(creation_options_param)

        extra_param = QgsProcessingParameterString(
            self.EXTRA,
            self.tr("Additional command-line parameters"),
            defaultValue=None,
            optional=True,
        )
        extra_param.setFlags(
            extra_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(extra_param)

        self.addParameter(
            QgsProcessingParameterRasterDestination(self.OUTPUT, self.tr("Output"))
        )

    def name(self):
        return "pansharp"

    def displayName(self):
        return self.tr("Pansharpening")

    def group(self):
        return self.tr("Raster miscellaneous")

    def groupId(self):
        return "rastermiscellaneous"

    def commandName(self):
        return "gdal_pansharpen"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        spectral = self.parameterAsRasterLayer(parameters, self.SPECTRAL, context)
        if spectral is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.SPECTRAL)
            )
        spectral_input_details = GdalUtils.gdal_connection_details_from_layer(spectral)

        panchromatic = self.parameterAsRasterLayer(
            parameters, self.PANCHROMATIC, context
        )
        if panchromatic is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.PANCHROMATIC)
            )
        panchromatic_input_details = GdalUtils.gdal_connection_details_from_layer(
            panchromatic
        )

        out = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, out)

        output_format = self.outputFormat(parameters, self.OUTPUT, context)
        if not output_format:
            raise QgsProcessingException(self.tr("Output format is invalid"))

        arguments = [
            panchromatic_input_details.connection_string,
            spectral_input_details.connection_string,
            out,
            "-r",
            self.methods[self.parameterAsEnum(parameters, self.RESAMPLING, context)][1],
            "-of",
            output_format,
        ]

        if panchromatic_input_details.credential_options:
            arguments.extend(
                panchromatic_input_details.credential_options_as_arguments()
            )

        options = self.parameterAsString(parameters, self.CREATION_OPTIONS, context)
        # handle backwards compatibility parameter OPTIONS
        if self.OPTIONS in parameters and parameters[self.OPTIONS] not in (None, ""):
            options = self.parameterAsString(parameters, self.OPTIONS, context)
        if options:
            arguments.extend(GdalUtils.parseCreationOptions(options))

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        return [
            self.commandName() + (".bat" if isWindows() else ".py"),
            GdalUtils.escapeAndJoin(arguments),
        ]
